import importlib
from pathlib import Path

from dash import Dash


def collect_callbacks(app: Dash):
    for file in Path(__file__).parent.glob("*.py"):
        if file.stem != "__init__":
            module = importlib.import_module(f".{file.stem}", package=__package__)
            if hasattr(module, "callbacks"):
                module.callbacks(app)



# https://www.ncbi.nlm.nih.gov/research/pubtator3/api
import json
import logging
import sys
import time
from collections.abc import Sequence
from datetime import datetime
from queue import Queue
from typing import Literal

import requests
from requests import Response, Session
from tqdm.auto import tqdm

from netmedex.biocjson_parser import biocjson_to_pubtator
from netmedex.exceptions import EmptyInput, NoArticles, UnsuccessfulRequest

# API GET limit: 100
PMID_REQUEST_SIZE = 100
# Fall back to "search" if "cite" failed
FALLBACK_SEARCH = True

# users post no more than three requests per second
# https://www.ncbi.nlm.nih.gov/research/pubtator3/api
SLEEP = 0.5

# Full text annotation is only availabe in `biocxml` and `biocjson` formats
# RESPONSE_FORMAT = ["pubtator", "biocxml", "biocjson"][2]
logger = logging.getLogger(__name__)


class PubTatorAPI:
    def __init__(
        self,
        query: str | None,
        pmid_list: Sequence[str] | None,
        savepath: str | None,
        search_type: Literal["query", "pmids"],
        sort: Literal["score", "date"],
        max_articles: int,
        use_mesh: bool,
        full_text: bool,
        debug: bool,
        queue: Queue | None = None,
    ):
        self.query = query
        self.pmid_list = pmid_list
        self.savepath = savepath
        self.search_type = search_type
        self.sort = sort
        self.max_articles = max_articles
        self.use_mesh = use_mesh
        self.full_text = full_text
        self.debug = debug
        self.queue = queue
        self.api_method = "cite" if sort == "date" else "search"
        self.return_progress = isinstance(queue, Queue)

    def run(self):
        if self.search_type == "query":
            if self.query is None or self.query.strip() == "":
                raise EmptyInput
            self.pmid_list = self.get_query_results()
        elif self.search_type == "pmids":
            if not self.pmid_list:
                raise EmptyInput

        if not self.pmid_list:
            raise NoArticles

        search_results = self.batch_publication_search()

        if self.use_mesh or self.full_text:
            retain_ori_text = False if self.use_mesh else True
            search_results = [
                biocjson_to_pubtator(
                    articles, retain_ori_text=retain_ori_text, role_type="identifier"
                )
                for articles in search_results
            ]

        self._write_results(search_results)

    def get_query_results(self):
        logger.info(f"Query: {self.query}")
        if self.api_method == "search":
            article_list = self._get_by_search()
        elif self.api_method == "cite":
            article_list = self._get_by_cite()

        return article_list

    def _get_by_search(self):
        res = send_search_query(self.query, type="search")

        article_list = []
        if not request_successful(res):
            return article_list

        res_json = res.json()
        total_articles = int(res_json["count"])
        page_size = int(res_json["page_size"])
        article_list.extend(get_article_ids(res_json))

        n_articles_to_request = get_n_articles(self.max_articles, total_articles)

        # Get search results in different pages until the max_articles is reached
        logger.info("Obtaining article PMIDs...")
        current_page = 1
        with Session() as session:
            with tqdm(total=n_articles_to_request, file=sys.stdout) as pbar:
                while current_page * page_size < n_articles_to_request:
                    current_page += 1
                    res = send_search_query_with_page(self.query, current_page, self.sort, session)
                    if request_successful(res):
                        article_list.extend(get_article_ids(res.json()))
                    pbar.update(page_size)
                    if self.return_progress:
                        self.queue.put(progress_message("search-search", pbar.n, pbar.total))
                pbar.n = pbar.total

        return article_list[:n_articles_to_request]

    def _get_by_cite(self):
        logger.info("Obtaining article PMIDs...")
        res = send_search_query(self.query, type="cite")
        if not request_successful(res):
            if FALLBACK_SEARCH:
                logger.warning("Fetching articles by 'cite' method failed. Switch to 'search'.")
                return self._get_by_search()
            else:
                unsuccessful_query(res.status_code)

        pmid_list = parse_cite_response(res.text)
        n_articles_to_request = get_n_articles(self.max_articles, len(pmid_list))
        if self.return_progress:
            self.queue.put(
                progress_message("search-cite", n_articles_to_request, n_articles_to_request)
            )

        return pmid_list[:n_articles_to_request]

    def batch_publication_search(self):
        results = []
        format = "biocjson" if self.use_mesh or self.full_text else "pubtator"
        with Session() as session:
            with tqdm(total=len(self.pmid_list), file=sys.stdout) as pbar:
                for start in range(0, len(self.pmid_list), PMID_REQUEST_SIZE):
                    end = start + PMID_REQUEST_SIZE
                    end = end if end < len(self.pmid_list) else None
                    res = send_publication_query(
                        pmid_string=",".join(self.pmid_list[start:end]),
                        article_id_type="pmids",
                        format=format,
                        full_text=self.full_text,
                        session=session,
                    )
                    if request_successful(res):
                        results.extend(self._append_json_or_text(res))
                    if end is not None:
                        pbar.update(PMID_REQUEST_SIZE)
                    else:
                        pbar.n = len(self.pmid_list)

                    if self.return_progress:
                        self.queue.put(progress_message("get", pbar.n, pbar.total))

        if self.return_progress:
            self.queue.put(None)

        if self.debug:
            now = datetime.now().strftime("%y%m%d%H%M%S")
            with open(f"./pubtator3_response_{now}.txt", "w") as f:
                f.writelines([json.dumps(o) + "\n" for o in results])
            with open(f"./pubtator3_pmid_{now}.txt", "w") as f:
                f.writelines([f"{id}\n" for id in self.pmid_list])
        return results

    def _append_json_or_text(self, res: Response):
        if self.full_text or self.use_mesh:
            if len(res.text.split("\n")) > 2:
                content = [json.loads(text) for text in res.text.split("\n")[:-1]]
            else:
                content = [res.json()]
        else:
            content = [res.text]

        return content

    def _write_results(self, results):
        if self.savepath is None:
            return

        header = []

        if self.use_mesh:
            header.append("##USE-MESH-VOCABULARY")
        if len(header) > 0:
            header.append("\n")

        with open(self.savepath, "w") as f:
            f.writelines(header)
            f.writelines(results)
            logger.info(f"Save to {self.savepath}")


def send_search_query(
    query: str,
    type: Literal["search", "cite"],
    session: Session | None = None,
):
    if type == "search":
        url = "https://www.ncbi.nlm.nih.gov/research/pubtator3-api/search/"
    elif type == "cite":
        url = "https://www.ncbi.nlm.nih.gov/research/pubtator3-api/cite/tsv"
    res = handle_session_get(url, params={"text": query}, session=session)
    time.sleep(SLEEP)
    return res


def send_search_query_with_page(
    query: str,
    page: int,
    sort: Literal["score", "date"],
    session: Session | None = None,
):
    url = "https://www.ncbi.nlm.nih.gov/research/pubtator3-api/search/"
    if sort == "score":
        params = {"text": query, "sort": "score desc", "page": page}
    elif sort == "date":
        params = {"text": query, "sort": "date desc", "page": page}
    res = handle_session_get(url, params, session)
    time.sleep(SLEEP)
    return res


def send_publication_query(
    pmid_string: str,
    article_id_type: Literal["pmids", "pmcids"],
    format: Literal["biocjson", "pubtator"],
    full_text: bool,
    session: Session | None = None,
):
    url = f"https://www.ncbi.nlm.nih.gov/research/pubtator3-api/publications/export/{format}"
    params = {article_id_type: pmid_string}
    if full_text:
        params["full"] = "true"
    res = handle_session_get(url, params, session)
    time.sleep(SLEEP)
    return res


def get_n_articles(max_articles: int, total_articles: int):
    logger.info(f"Find {total_articles} articles")
    n_articles_to_request = max_articles if total_articles > max_articles else total_articles
    logger.info(f"Requesting {n_articles_to_request} articles...")
    return n_articles_to_request


def get_article_ids(res_json):
    return [str(article.get("pmid")) for article in res_json["results"]]


def parse_cite_response(res_text: str):
    pmid_list = []
    for line in res_text.split("\n"):
        if line.startswith("#") or line == "":
            continue
        # [pmid, title, journal]
        pmid = line.split("\t")[0]
        pmid_list.append(pmid)
    return pmid_list


def request_successful(res: Response):
    if res.status_code != 200:
        logger.info("Unsuccessful request")
        logger.debug(f"Response status code: {res.status_code}")
        return False
    return True


def unsuccessful_query(status_code: int):
    if status_code == 502:
        msg = "Possibly too many articles. Please try more specific queries."
    else:
        msg = "Please retry later."

    raise UnsuccessfulRequest(msg)


def handle_session_get(url: str, params, session: Session | None = None):
    try:
        res = session.get(url, params=params)
    except Exception:
        res = requests.get(url, params=params)

    return res


def progress_message(status, progress, total):
    return f"{status}/{progress}/{total}"



import dash_bootstrap_components as dbc
from dash import Input, Output, dcc, html

from webapp.utils import display


def generate_query_component(hidden=False):
    return html.Div(
        [
            html.H5("Query"),
            dbc.Input(
                placeholder="Enter a query (ex: dimethylnitrosamine)", type="text", id="data-input"
            ),
        ],
        hidden=hidden,
    )


def generate_pmid_component(hidden=False):
    return html.Div(
        [
            html.H5("PMID"),
            dbc.Input(
                placeholder="Enter PMIDs (ex: 33422831,33849366)", type="text", id="data-input"
            ),
        ],
        hidden=hidden,
    )


def generate_pmid_file_component(hidden=False):
    return html.Div(
        [
            html.H5("PMID File"),
            dcc.Upload(
                id="pmid-file-data",
                children=html.Div(
                    ["Drag and Drop or ", html.A("Select Files", className="hyperlink")],
                    className="upload-box form-control",
                ),
            ),
            html.Div(id="output-data-upload"),
        ],
        hidden=hidden,
    )


def callbacks(app):
    @app.callback(
        Output("api-wrapper", "style"),
        Output("pubtator-file-wrapper", "style"),
        Input("api-toggle-items", "value"),
        prevent_initial_call=True,
    )
    def update_api_toggle(api_toggle):
        if api_toggle == "api":
            return display.block, display.none
        elif api_toggle == "file":
            return display.none, display.block

    @app.callback(
        Output("input-type", "children"),
        Input("input-type-selection", "value"),
    )
    def update_input_type(input_type):
        if input_type == "query":
            return [
                generate_query_component(hidden=False),
                generate_pmid_file_component(hidden=True),
            ]
        elif input_type == "pmids":
            return [
                generate_pmid_component(hidden=False),
                generate_pmid_file_component(hidden=True),
            ]
        elif input_type == "pmid_file":
            return [
                generate_query_component(hidden=True),
                generate_pmid_file_component(hidden=False),
            ]



from dotenv import load_dotenv

load_dotenv()

import os

import dash_bootstrap_components as dbc
import diskcache
from dash import ClientsideFunction, Dash, Input, Output, dcc, html
from dash.long_callback import DiskcacheLongCallbackManager

from netmedex.utils import config_logger
from webapp.callbacks import collect_callbacks
from webapp.utils import clean_up_files

config_logger(is_debug=(os.getenv("LOGGING_DEBUG") == "true"))


cache = diskcache.Cache("./cache")
long_callback_manager = DiskcacheLongCallbackManager(cache)

app = Dash(
    __name__,
    external_stylesheets=[dbc.themes.BOOTSTRAP],
    long_callback_manager=long_callback_manager,
    suppress_callback_exceptions=True,
)
app.title = "NetMedEx"
app._favicon = "NetMedEx_ico.ico"

from webapp.components.graph import graph
from webapp.components.sidebar import sidebar

current_session_path = dcc.Store(id="current-session-path")

content = html.Div(
    [current_session_path, sidebar, graph],
    className="d-flex flex-row position-relative h-100",
)

app.layout = html.Div([content, html.Div(id="post-js-scripts")], id="main-container")


def main():
    try:
        collect_callbacks(app)
        app.clientside_callback(
            ClientsideFunction(namespace="clientside", function_name="info_scroll"),
            Output("post-js-scripts", "children"),
            Input("post-js-scripts", "id"),
        )
        app.run(host=os.getenv("HOST", "127.0.0.1"), port=os.getenv("PORT", 8050))
    finally:
        clean_up_files()


if __name__ == "__main__":
    main()



import dash_bootstrap_components as dbc
from dash import dcc, html

from webapp.components.advanced_settings import advanced_settings
from webapp.components.utils import generate_param_title
from webapp.utils import display

api_or_file = html.Div(
    [
        html.Div(
            [
                generate_param_title(
                    "Source",
                    (
                        "PubTator3 API: Search + Network Generation\n"
                        "PubTator File: Network Generation from PubTator File"
                    ),
                ),
                dbc.RadioItems(
                    id="api-toggle-items",
                    options=[
                        {"label": "PubTator3 API", "value": "api"},
                        {"label": "PubTator File", "value": "file"},
                    ],
                    value="api",
                    inline=True,
                ),
            ],
            className="param",
        )
    ],
    id="api-toggle",
)

pubtator_file = html.Div(
    [
        html.Div(
            [
                generate_param_title(
                    "PubTator File",
                    "The file downloaded using the 'PubTator File' button after running the 'PubTator3 API'",
                ),
                dcc.Upload(
                    id="pubtator-file-data",
                    children=html.Div(
                        ["Drag and Drop or ", html.A("Select Files", className="hyperlink")],
                        className="upload-box form-control",
                    ),
                ),
                html.Div(id="pubtator-file-upload"),
            ],
            className="param",
        )
    ],
    style=display.none,
    id="pubtator-file-wrapper",
)

api_params = html.Div(
    [
        html.Div(
            [
                generate_param_title(
                    "Search Type",
                    (
                        "Text Search: Use keywords to retrieve relevant articles (use double quotes to match whole words and AND/OR to combine keywords)\n"
                        "PMID: Retrieve articles by PubMed Identifier (PMID)\n"
                        "PMID File: Retrieve articles by a text file of PMIDs (one per line)"
                    ),
                ),
                dcc.Dropdown(
                    id="input-type-selection",
                    options=[
                        {"label": "Text Search", "value": "query"},
                        {"label": "PMID", "value": "pmids"},
                        {"label": "PMID File", "value": "pmid_file"},
                    ],
                    value="query",
                    style={"width": "200px"},
                ),
            ],
            className="param",
        ),
        html.Div(id="input-type", className="param"),
        html.Div(
            [
                generate_param_title("Sort", "Sort articles by recency or relevance"),
                dbc.RadioItems(
                    id="sort-toggle-methods",
                    options=[
                        {"label": "Recency", "value": "date"},
                        {"label": "Relevance", "value": "score"},
                    ],
                    value="date",
                    inline=True,
                ),
            ],
            className="param",
        ),
        html.Div(
            [
                generate_param_title(
                    "PubTator3 Parameters",
                    (
                        "Use MeSH Vocabulary: Replace original text in articles with standardized MeSH terms\n"
                        "Full Text: Build network from full-text articles if available, defaulting to abstracts otherwise (not recommended to enable)"
                    ),
                ),
                dbc.Checklist(
                    options=[
                        {"label": "Use MeSH Vocabulary", "value": "use_mesh"},
                        {"label": "Full Text", "value": "full_text"},
                    ],
                    switch=True,
                    id="pubtator-params",
                    value=["use_mesh"],
                ),
            ],
            className="param",
        ),
    ],
    id="api-wrapper",
)


network_params = html.Div(
    [
        html.Div(
            [
                generate_param_title(
                    "Node Filter",
                    (
                        "All: Retain all annotations\n"
                        "MeSH: Retain annotations with standardized MeSH terms only\n"
                        "BioREx Relation: Retain annotations with high-confidence relationships from PubTator3 BioREx model"
                    ),
                ),
                dcc.Dropdown(
                    id="node-type",
                    options=[
                        {"label": "All", "value": "all"},
                        {"label": "MeSH", "value": "mesh"},
                        {"label": "BioREx Relation", "value": "relation"},
                    ],
                    value="all",
                    style={"width": "200px"},
                ),
            ],
            className="param",
        ),
        html.Div(
            [
                generate_param_title(
                    "Weighting Method",
                    (
                        "Frequency: Calculate edge weights using co-occurence counts\n"
                        "NPMI: Calulate edge weights using normalized mutual pointwise information"
                    ),
                ),
                dcc.Dropdown(
                    id="weighting-method",
                    options=[
                        {"label": "Frequency", "value": "freq"},
                        {"label": "NPMI", "value": "npmi"},
                    ],
                    value="freq",
                    style={"width": "200px"},
                ),
            ],
            className="param",
        ),
        html.Div(
            [
                generate_param_title(
                    "Edge Weight Cutoff",
                    "Set the minimum edge weight to filter the graph",
                ),
                dcc.Slider(
                    0,
                    20,
                    1,
                    value=2,
                    marks=None,
                    id="cut-weight",
                    tooltip={"placement": "bottom", "always_visible": True},
                ),
            ],
            className="param",
        ),
        html.Div(
            [
                generate_param_title(
                    "Network Parameters",
                    "Community: Group nodes into communities",
                ),
                dbc.Checklist(
                    options=[
                        {"label": "Community", "value": "community"},
                    ],
                    switch=True,
                    id="cy-params",
                    value=["community"],
                ),
            ],
            className="param",
        ),
    ],
    id="cy-wrapper",
)


progress = html.Div(
    [
        html.Div(
            [html.H5("Progress"), html.P(" ", id="progress-status"), dbc.Progress(id="progress")],
            className="param",
        ),
        html.Div([dbc.Button("Submit", id="submit-button"), html.Div(id="output")]),
    ],
    id="progress-wrapper",
)


sidebar = html.Div(
    [advanced_settings, api_or_file, api_params, pubtator_file, network_params, progress],
    className="sidebar",
)



import math


def normalized_pointwise_mutual_information(
    n_x: float,
    n_y: float,
    n_xy: float,
    N: int,
    n_threshold: int,
    below_threshold_default: float,
):
    if n_xy == 0:
        npmi = -1
    elif (n_xy / N) == 1:
        npmi = 1
    else:
        npmi = -1 + (math.log2(n_x / N) + math.log2(n_y / N)) / math.log2(n_xy / N)

    # non-normalized
    # pmi = math.log2(p_x) + math.log2(p_y) - math.log2(p_xy)

    if n_x < n_threshold or n_y < n_threshold:
        npmi = min(npmi, below_threshold_default)

    return npmi



import json
import re
from typing import Literal

import networkx as nx

from netmedex.cytoscape_html_template import HTML_TEMPLATE

SHAPE_JS_MAP = {"PARALLELOGRAM": "RHOMBOID"}
COMMUNITY_NODE_PATTERN = re.compile(r"^c\d+$")


def save_as_html(G: nx.Graph, savepath: str, layout="preset"):
    with open(savepath, "w") as f:
        cytoscape_js = create_cytoscape_js(G, style="cyjs")
        f.write(HTML_TEMPLATE.format(cytoscape_js=json.dumps(cytoscape_js), layout=layout))


def save_as_json(G: nx.Graph, savepath: str):
    with open(savepath, "w") as f:
        cytoscape_js = create_cytoscape_js(G, style="dash")
        f.write(json.dumps(cytoscape_js))


def create_cytoscape_js(G: nx.Graph, style: Literal["dash", "cyjs"] = "cyjs"):
    # TODO: Check whether to set id for edges
    with_id = False
    nodes = [create_cytoscape_node(node) for node in G.nodes(data=True)]
    edges = [create_cytoscape_edge(edge, G, with_id) for edge in G.edges(data=True)]

    if style == "cyjs":
        elements = nodes + edges
    elif style == "dash":
        elements = {"elements": {"nodes": nodes, "edges": edges}}

    return elements


def create_cytoscape_node(node):
    def convert_shape(shape):
        return SHAPE_JS_MAP.get(shape, shape).lower()

    node_id, node_attr = node

    node_info = {
        "data": {
            "id": node_attr["_id"],
            "parent": node_attr.get("parent", None),
            "color": node_attr["color"],
            "label_color": node_attr["label_color"],
            "label": node_attr["name"],
            "shape": convert_shape(node_attr["shape"]),
        },
        "position": {
            "x": round(node_attr["pos"][0], 3),
            "y": round(node_attr["pos"][1], 3),
        },
    }

    # Community nodes
    if COMMUNITY_NODE_PATTERN.search(node_attr["_id"]):
        node_info["classes"] = "top-center"

    return node_info


def create_cytoscape_edge(edge, G, with_id=True):
    node_id_1, node_id_2, edge_attr = edge

    edge_info = {
        "data": {
            "source": G.nodes[node_id_1]["_id"],
            "target": G.nodes[node_id_2]["_id"],
            "label": f"{G.nodes[node_id_1]['name']} (interacts with) {G.nodes[node_id_2]['name']}",
            "weight": round(max(float(edge_attr["edge_width"]), 1), 1),
            "pmids": edge_attr["pmids"],
            "edge_type": edge_attr["type"],
        }
    }

    if with_id:
        edge_info["data"]["id"] = edge_attr["_id"]

    return edge_info



import uuid

import dash_cytoscape as cyto
from dash import ClientsideFunction, Input, Output, State, clientside_callback, no_update

from netmedex.cytoscape_js import create_cytoscape_js
from webapp.callbacks.graph_utils import rebuild_graph


def generate_cytoscape_js_network(graph_layout, graph_json):
    if graph_json is not None:
        elements = [*graph_json["elements"]["nodes"], *graph_json["elements"]["edges"]]
    else:
        elements = []

    cytoscape_graph = cyto.Cytoscape(
        id="cy",
        minZoom=0.1,
        maxZoom=20,
        wheelSensitivity=0.3,
        style={},
        layout={"name": graph_layout},
        stylesheet=[
            {
                "selector": "node",
                "style": {
                    "text-valign": "center",
                    "label": "data(label)",
                    "shape": "data(shape)",
                    "color": "data(label_color)",
                    "background-color": "data(color)",
                },
            },
            {
                "selector": ":parent",
                "style": {
                    "background-opacity": 0.3,
                },
            },
            {
                "selector": "edge",
                "style": {
                    "width": "data(weight)",
                },
            },
            {
                "selector": ".top-center",
                "style": {
                    "text-valign": "top",
                    "text-halign": "center",
                    "font-size": "20px",
                },
            },
        ],
        elements=elements,
    )

    return cytoscape_graph


# TODO: temporary workaround for cannot create edge for non-existence source or target
# https://github.com/plotly/dash-cytoscape/issues/106
def generate_new_id(graph_json):
    id_map = {}
    # Give new id to each node
    for node in graph_json["elements"]["nodes"]:
        new_id = str(uuid.uuid4())
        old_id = node["data"]["id"]
        id_map[old_id] = new_id
        node["data"]["id"] = new_id

    # Update parent node id
    for node in graph_json["elements"]["nodes"]:
        if (old_parent_id := node["data"]["parent"]) is not None:
            node["data"]["parent"] = id_map[old_parent_id]

    # Update source and target for each edge
    for edge in graph_json["elements"]["edges"]:
        edge["data"]["source"] = id_map[edge["data"]["source"]]
        edge["data"]["target"] = id_map[edge["data"]["target"]]

    return graph_json


def callbacks(app):
    @app.callback(
        Output("progress", "value"),
        Output("progress", "max"),
        Output("progress", "label"),
        Output("progress-status", "children"),
        Input("cy-graph", "children"),
        State("progress-status", "children"),
        running=[(Input("submit-button", "disabled"), True, False)],
        prevent_initial_call=True,
    )
    def plot_cytoscape_graph(graph_children, progress):
        if graph_children is not None:
            return 1, 1, "1/1", "Done"
        else:
            return 1, 1, "", progress

    @app.callback(
        Output("cy-graph", "children"),
        Output("is-new-graph", "data", allow_duplicate=True),
        Output("memory-node-degree", "data"),
        Output("memory-graph-cut-weight", "data", allow_duplicate=True),
        Input("node-degree", "value"),
        Input("graph-cut-weight", "value"),
        State("memory-node-degree", "data"),
        State("memory-graph-cut-weight", "data"),
        State("cy-graph-container", "style"),
        State("graph-layout", "value"),
        State("is-new-graph", "data"),
        State("current-session-path", "data"),
        prevent_initial_call=True,
    )
    def update_graph(
        new_node_degree,
        new_cut_weight,
        old_node_degree,
        old_cut_weight,
        container_style,
        graph_layout,
        is_new_graph,
        savepath,
    ):
        if container_style["visibility"] == "hidden":
            cy_graph = generate_cytoscape_js_network(graph_layout, None)
            return cy_graph, False, new_node_degree, new_cut_weight

        if new_node_degree is None:
            new_node_degree = old_node_degree

        conditions = (
            is_new_graph or new_cut_weight != old_cut_weight or new_node_degree != old_node_degree
        )
        if conditions:
            G = rebuild_graph(
                new_node_degree,
                new_cut_weight,
                format="html",
                with_layout=True,
                graph_path=savepath["graph"],
            )
            graph_json = create_cytoscape_js(G, style="dash")
            graph_json = generate_new_id(graph_json)
            cy_graph = generate_cytoscape_js_network(graph_layout, graph_json)
            return cy_graph, False, new_node_degree, new_cut_weight
        else:
            return no_update, False, new_node_degree, new_cut_weight

    @app.callback(
        Output("cy", "layout"),
        Output("cy", "elements", allow_duplicate=True),
        Input("graph-layout", "value"),
        State("node-degree", "value"),
        State("graph-cut-weight", "value"),
        State("cy", "elements"),
        State("current-session-path", "data"),
        prevent_initial_call=True,
    )
    def update_graph_layout(layout, node_degree, weight, elements, savepath):
        if layout == "preset":
            G = rebuild_graph(
                node_degree, weight, format="html", with_layout=True, graph_path=savepath["graph"]
            )
            graph_json = create_cytoscape_js(G, style="dash")
            graph_json = generate_new_id(graph_json)
            elements = [*graph_json["elements"]["nodes"], *graph_json["elements"]["edges"]]

        return {"name": layout}, elements

    clientside_callback(
        ClientsideFunction(namespace="clientside", function_name="show_edge_info"),
        Output("edge-info-container", "style"),
        Output("edge-info", "children"),
        Input("cy", "selectedEdgeData"),
        State("cy", "tapEdgeData"),
        State("pmid-title-dict", "data"),
        prevent_initial_call=True,
    )



# PYTHON_ARGCOMPLETE_OK

import argparse
import logging
import sys
from pathlib import Path

from netmedex.utils import config_logger

logger = logging.getLogger(__name__)


def main():
    args = parse_args(sys.argv[1:])
    args.entry_func(args)


def pubtator_entry(args):
    from netmedex.exceptions import EmptyInput, NoArticles, UnsuccessfulRequest
    from netmedex.pubtator_core import PubTatorAPI
    from netmedex.pubtator_utils import create_savepath, load_pmids

    debug = args.debug
    logfile_name = "search" if debug else None
    config_logger(debug, logfile_name)

    num_inputs = sum(arg is not None for arg in [args.pmids, args.pmid_file, args.query])
    if num_inputs != 1:
        logger.info("Please specify only one of the following: --query, --pmids, --pmid_file")
        sys.exit()

    query = None
    pmid_list = None
    if args.query is not None:
        search_type = "query"
        query = args.query
    elif args.pmids is not None:
        search_type = "pmids"
        pmid_list = load_pmids(args.pmids, load_from="string")
        logger.info(f"Find {len(pmid_list)} PMIDs")
    elif args.pmid_file is not None:
        search_type = "pmids"
        logger.info(f"Load PMIDs from: {args.pmid_file}")
        pmid_list = load_pmids(args.pmid_file, load_from="file")
        logger.info(f"Find {len(pmid_list)} PMIDs")

    if search_type == "query":
        suffix = query.replace(" ", "_")
    if search_type == "pmids":
        if pmid_list:
            suffix = f"{pmid_list[0]}_total_{len(pmid_list)}"
        else:
            suffix = ""
    savepath = create_savepath(args.output, type=search_type, suffix=suffix)

    pubtator_api = PubTatorAPI(
        query=query,
        pmid_list=pmid_list,
        savepath=str(savepath),
        search_type=search_type,
        sort=args.sort,
        max_articles=args.max_articles,
        full_text=args.full_text,
        use_mesh=args.use_mesh,
        debug=args.debug,
        queue=None,
    )

    try:
        pubtator_api.run()
    except (NoArticles, EmptyInput, UnsuccessfulRequest) as e:
        logger.error(str(e))


def network_entry(args):
    from netmedex.network_core import NetworkBuilder

    debug = args.debug
    logfile_name = "network" if debug else None
    config_logger(debug, logfile_name)

    pubtator_filepath = Path(args.input)
    if args.output is None:
        savepath = pubtator_filepath.with_suffix(f".{args.format}")
    else:
        savepath = Path(args.output)
        savepath.parent.mkdir(parents=True, exist_ok=True)

    network_builder = NetworkBuilder(
        pubtator_filepath=str(pubtator_filepath),
        savepath=str(savepath),
        node_type=args.node_type,
        output_filetype=args.format,
        weighting_method=args.weighting_method,
        edge_weight_cutoff=args.cut_weight,
        pmid_weight_filepath=args.pmid_weight,
        max_edges=args.max_edges,
        community=args.community,
        debug=args.debug,
    )

    network_builder.run()


def webapp_entry(args):
    from webapp.app import main

    main()


def parse_args(args):
    parser = argparse.ArgumentParser()
    subparser = parser.add_subparsers()

    pubtator_subparser = subparser.add_parser(
        "search",
        parents=[get_pubtator_parser()],
        help="Search PubMed articles and obtain annotations",
    )
    pubtator_subparser.set_defaults(entry_func=pubtator_entry)

    network_subparser = subparser.add_parser(
        "network",
        parents=[get_network_parser()],
        help="Build a network from annotations",
    )
    network_subparser.set_defaults(entry_func=network_entry)

    webapp_subparser = subparser.add_parser(
        "run",
        help="Run NetMedEx app",
    )
    webapp_subparser.set_defaults(entry_func=webapp_entry)

    return parser.parse_args(args)


def get_pubtator_parser():
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "-q",
        "--query",
        default=None,
        help="Query string",
    )
    parser.add_argument(
        "-o",
        "--output",
        default=None,
        help="Output path (default: [CURRENT_DIR].pubtator)",
    )
    parser.add_argument(
        "-p",
        "--pmids",
        default=None,
        type=str,
        help="PMIDs for the articles (comma-separated)",
    )
    parser.add_argument(
        "-f",
        "--pmid_file",
        default=None,
        help="Filepath to load PMIDs (one per line)",
    )
    parser.add_argument(
        "-s",
        "--sort",
        default="date",
        choices=["score", "date"],
        help="Sort articles in descending order by (default: date)",
    )
    parser.add_argument(
        "--max_articles",
        type=int,
        default=1000,
        help="Maximal articles to request from the searching result (default: 1000)",
    )
    parser.add_argument(
        "--full_text",
        action="store_true",
        help="Collect full-text annotations if available",
    )
    parser.add_argument(
        "--use_mesh",
        action="store_true",
        help="Use MeSH vocabulary instead of the most commonly used original text in articles",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Print debug information",
    )

    return parser


def get_network_parser():
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        help="Path to the pubtator file",
    )
    parser.add_argument(
        "-o",
        "--output",
        default=None,
        help="Output path (default: [INPUT_DIR].[FORMAT_EXT])",
    )
    parser.add_argument(
        "-w",
        "--cut_weight",
        type=int,
        default=2,
        help="Discard the edges with weight smaller than the specified value (default: 2)",
    )
    parser.add_argument(
        "-f",
        "--format",
        choices=["xgmml", "html", "json"],
        default="html",
        help="Output format (default: html)",
    )
    parser.add_argument(
        "--node_type",
        choices=["all", "mesh", "relation"],
        default="all",
        help="Keep specific types of nodes (default: all)",
    )
    parser.add_argument(
        "--weighting_method",
        choices=["freq", "npmi"],
        default="freq",
        help="Weighting method for network edge (default: freq)",
    )
    parser.add_argument(
        "--pmid_weight",
        default=None,
        help="CSV file for the weight of the edge from a PMID (default: 1)",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Print debug information",
    )
    parser.add_argument(
        "--community",
        action="store_true",
        help="Divide nodes into distinct communities by the Louvain method",
    )
    parser.add_argument(
        "--max_edges",
        type=int,
        default=0,
        help="Maximum number of edges to display (default: 0, no limit)",
    )

    return parser



import logging
import sys
from datetime import datetime
from uuid import uuid4


def generate_uuid():
    return str(uuid4())


def config_logger(is_debug, filename=None):
    handlers = [logging.StreamHandler(stream=sys.stdout)]

    if filename is not None:
        now = datetime.now().strftime("%y%m%d%H%M%S")
        logfile = f"{filename}_{now}.log"
        handlers.append(logging.FileHandler(logfile, mode="w"))

    if is_debug:
        logging.basicConfig(
            format="%(asctime)s [%(levelname)s] %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            level=logging.DEBUG,
            handlers=handlers,
        )
    else:
        logging.basicConfig(format="%(message)s", level=logging.INFO, handlers=handlers)



from pathlib import Path
from typing import Literal


def load_pmids(input_data, load_from: Literal["string", "file"]):
    if input_data is None:
        return []
    input_data = str(input_data).strip()
    if load_from == "string":
        pmids = input_data.split(",")
    elif load_from == "file":
        pmids = []
        with open(input_data) as f:
            for line in f.readlines():
                pmids.extend(line.strip().split(","))

    pmids = drop_if_not_num(pmids)

    return pmids


def drop_if_not_num(id_list):
    checked_list = []
    for id in id_list:
        id = id.strip()
        try:
            _ = int(id)
            checked_list.append(id)
        except ValueError:
            pass

    return checked_list


def create_savepath(output_path, type, suffix):
    if output_path is None:
        if type == "query":
            savepath = Path(f"./query_{suffix}.pubtator")
        elif type == "pmids":
            savepath = Path(f"./pmids_{suffix}.pubtator")
    else:
        savepath = Path(output_path)
        savepath.parent.mkdir(parents=True, exist_ok=True)

    return savepath



HTML_TEMPLATE = """<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="UTF-8">
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
  <title>Cytoscape Network</title>
</head>
<style>
  * {{
    padding: 0;
    margin: 0;
    box-sizing: border-box;
  }}

  body {{
    background-color: #eeeeee;
  }}

  #cy {{
      width: 98%;
      height: 98%;
      position: absolute;
  }}

  #legend-container {{
      position: absolute;
      padding: 5px;
      background-color: #eeeeee;
      border: 1px solid #585858;
      bottom: 50px;
      right: 50px;
      z-index: 100;
  }}

  .legend-box {{
    display: grid;
    grid-template-columns: 1fr 2fr;
    align-items: center;
    justify-items: center;
    width: 130px;
  }}

  .legend-box p {{
      margin: 0;
      font-size: 12px;
  }}
</style>
<body>
<div id="cy"></div>
<div id="legend-container">
  <div class="legend-box">
    <svg width="25px" height="25px" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 120 120"><defs><style>.cls-1{{fill:#fd8d3c;}}</style></defs><rect class="cls-1" x="22.5" y="22.5" width="75" height="75" transform="translate(60 -24.85) rotate(45)"/></svg>
    <p>Species</p>
  </div>
  <div class="legend-box">
    <svg width="25px" height="25px" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 120 120"><defs><style>.cls-2{{fill:#67a9cf;}}</style></defs><circle class="cls-2" cx="60" cy="60" r="45"/></svg>
    <p>Chemical</p>
  </div>
  <div class="legend-box">
    <svg width="25px" height="25px" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 120 120"><defs><style>.cls-3{{fill:#74c476;}}</style></defs><polygon class="cls-3" points="60 20 13.81 100 106.19 100 60 20"/></svg>
    <p>Gene</p>
  </div>
  <div class="legend-box">
    <svg width="25px" height="25px" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 120 120"><defs><style>.cls-4{{fill:#8c96c6;}}</style></defs><rect class="cls-4" x="17.5" y="17.5" width="85" height="85" rx="23.84"/></svg>
    <p>Disease</p>
  </div>
  <div class="legend-box">
    <svg width="25px" height="25px" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 120 120"><defs><style>.cls-5{{fill:#bdbdbd;}}</style></defs><polygon class="cls-5" points="60.17 102.5 109.08 17.5 60.17 39.19 10.93 17.5 60.17 102.5"/></svg>
    <p>CellLine</p>
  </div>
  <div class="legend-box">
    <svg width="25px" height="25px" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 120 120"><defs><style>.cls-6{{fill:#fccde5;}}</style></defs><polygon class="cls-6" points="104.5 110 37.83 110 15.5 10 82.17 10 104.5 110"/></svg>
    <p>DNAMutation</p>
  </div>
  <div class="legend-box">
    <svg width="25px" height="25px" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 120 120"><defs><style>.cls-7{{fill:#fa9fb5;}}</style></defs><polygon class="cls-7" points="85 16.7 35 16.7 10 60 35 103.3 85 103.3 110 60 85 16.7"/></svg>
    <p>ProteinMutation</p>
  </div>
  <div class="legend-box">
    <svg width="25px" height="25px" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 120 120"><defs><style>.cls-8{{fill:#ffffb3;}}</style></defs><polygon class="cls-8" points="79.13 13.81 40.87 13.81 13.81 40.87 13.81 79.13 40.87 106.19 79.13 106.19 106.19 79.13 106.19 40.87 79.13 13.81"/></svg>
    <p>SNP</p>
  </div>
</div>
</body>
<script src="https://cdnjs.cloudflare.com/ajax/libs/cytoscape/3.30.1/cytoscape.min.js"></script>
<script>
  let cy = cytoscape({{
    container: document.getElementById("cy"),
    elements: {cytoscape_js},
    layout: {{"name": "{layout}"}},
    style: [
    {{
      selector: "node",
      style: {{
        "text-valign": "center",
        "label": "data(label)",
        "shape": "data(shape)",
        "color": "data(label_color)",
        "background-color" : "data(color)",
      }},
    }},
    {{
      selector: ":parent",
      style: {{
        "background-opacity": 0.3,
      }},
    }},
    {{
      selector: "edge",
      style: {{
        "width": "data(weight)",
      }},
    }},
    {{
      selector: ".top-center",
      style: {{
        "text-valign": "top",
        "text-halign": "center",
        "font-size": "20px",
      }},
    }}
  ]
  }})
</script>
</html>
"""



import logging
import re
from collections import defaultdict
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import Literal

from netmedex.stemmers import s_stemmer
from netmedex.utils import generate_uuid

HEADER_SYMBOL = "##"
MUTATION_PATTERNS = {
    "tmvar": re.compile(r"(tmVar:[^;]+)"),
    "hgvs": re.compile(r"(HGVS:[^;]+)"),
    "rs": re.compile(r"(RS#:[^;]+)"),
    "variantgroup": re.compile(r"(VariantGroup:[^;]+)"),
    "gene": re.compile(r"(CorrespondingGene:[^;]+)"),
    "species": re.compile(r"(CorrespondingSpecies:[^;]+)"),
    "ca": re.compile(r"(CA#:[^;]+)"),
}

logger = logging.getLogger(__name__)


@dataclass
class PubTatorNodeData:
    id: str
    mesh: str
    type: str
    name: str | defaultdict[str, int]
    pmids: set[str]


@dataclass
class PubTatorEdgeData:
    id: str
    pmid: str
    relation: str | None = None


@dataclass
class PubTatorAnnotation:
    pmid: str
    start: str
    end: str
    name: str
    type: str
    mesh: str


@dataclass
class PubTatorRelation:
    pmid: str
    relation: str
    mesh1: str
    mesh2: str


class PubTatorLine:
    use_mesh = False

    def __init__(self, line: str):
        self.values: list[str] = [value.strip() for value in line.split("\t")]
        self.type: str = self.assign_line_type(self.values)
        self._data: PubTatorAnnotation | PubTatorRelation | None = None

    @staticmethod
    def assign_line_type(values: Sequence):
        if len(values) == 4:
            return "relation"
        elif len(values) == 6:
            return "annotation"
        else:
            return "unknown"

    @property
    def data(self):
        if self._data is None:
            self._data = self.parse_line()
        return self._data

    def parse_line(self):
        if self.type == "annotation":
            data = PubTatorAnnotation(*self.values)
        elif self.type == "relation":
            data = PubTatorRelation(*self.values)
        else:
            data = None
        return data

    def normalize_name(self):
        self.data.name = s_stemmer(self.data.name.lower())

    def parse_mesh(self, mesh_map: Mapping[str, str]):
        def convert_mesh(mesh: str, type_: str):
            converted = f"{type_}_{mesh}"
            if mesh not in mesh_map:
                mesh_map[mesh] = converted
            return converted

        anno_type = self.data.type
        mesh = self.data.mesh

        if not self.use_mesh:
            self.normalize_name()

        if anno_type in ("DNAMutation", "ProteinMutation", "SNP"):
            match_data = {}
            for key, pattern in MUTATION_PATTERNS.items():
                try:
                    match_data[key] = f"{re.search(pattern, mesh).group(1)};"
                except AttributeError:
                    match_data[key] = ""
            if anno_type == "DNAMutation":
                mesh = (
                    f'{match_data["gene"]}{match_data["species"]}{match_data["variantgroup"]}'
                    f'{match_data["tmvar"].split("|")[0]}'
                ).strip(";")
            elif anno_type == "ProteinMutation":
                mesh = f'{match_data["rs"]}{match_data["hgvs"]}{match_data["gene"]}'.strip(";")
            elif anno_type == "SNP":
                mesh = f'{match_data["rs"]}{match_data["hgvs"]}{match_data["gene"]}'.strip(";")
            result = [{"key": mesh, "mesh": mesh}]
        elif anno_type == "Gene":
            mesh_list = mesh.split(";")
            result = [{"key": convert_mesh(mesh, "gene"), "mesh": mesh} for mesh in mesh_list]
        elif anno_type == "Species":
            result = [{"key": convert_mesh(mesh, "species"), "mesh": mesh}]
        else:
            result = [{"key": mesh, "mesh": mesh}]

        return result


@dataclass
class PubTatorResult:
    node_dict: dict[str, PubTatorNodeData]
    edge_dict: dict[str, list[PubTatorEdgeData]]
    non_isolated_nodes: set[str]
    num_pmids: int
    pmid_title_map: dict[str, str]


class PubTatorParser:
    def __init__(
        self,
        pubtator_filepath: str,
        node_type: Literal["all", "mesh", "relation"],
    ):
        self.pubtator_file = pubtator_filepath
        self.node_type = node_type
        self._mesh_only: bool = True if node_type in ("mesh", "relation") else False
        self.num_pmids: int = 0
        self.mesh_map: dict[str, str] = {}
        self.pmid_title_map: dict[str, str] = {}
        self.node_dict_each: dict[str, PubTatorNodeData] = {}
        self.node_dict: dict[str, PubTatorNodeData] = {}
        self.edge_dict: defaultdict[str, list[PubTatorEdgeData]] = defaultdict(list)
        self.non_isolated_nodes: set[str] = set()

    def parse(self):
        pmid = -1
        last_pmid = -1
        node_dict_each = {}

        self._parse_header(self.pubtator_file)
        with open(self.pubtator_file) as f:
            for line in f.readlines():
                if self._is_title(line):
                    pmid = self._parse_title(line)
                    self.num_pmids += 1
                    continue
                if self.node_type == "relation":
                    parsed_line = PubTatorLine(line)
                    self._parse_line_relation(parsed_line)
                else:
                    if pmid != last_pmid:
                        self._create_complete_graph(node_dict_each, pmid)
                        last_pmid = pmid
                        node_dict_each = {}
                    parsed_line = PubTatorLine(line)
                    if parsed_line.type == "annotation":
                        self._add_node(parsed_line, node_dict_each)
            self._create_complete_graph(node_dict_each, pmid)
        self._determine_mesh_term_labels()
        self._get_non_isolated_nodes()

        return PubTatorResult(
            node_dict=self.node_dict,
            edge_dict=dict(self.edge_dict),
            non_isolated_nodes=self.non_isolated_nodes,
            num_pmids=self.num_pmids,
            pmid_title_map=self.pmid_title_map,
        )

    def _parse_line_relation(self, line: PubTatorLine):
        if line.type == "annotation":
            self._add_node(line, {})
        elif line.type == "relation":
            self._create_edges_for_relations(line)

    def _add_node(
        self,
        line: PubTatorLine,
        node_dict_each: dict[str, PubTatorNodeData],
    ):
        data = line.data
        if data.mesh in ("-", ""):
            if self._mesh_only:
                return
            # By Text
            line.normalize_name()

            name = data.name
            if not self._node_id_registered(line, name):
                self.node_dict[name] = PubTatorNodeData(
                    id=generate_uuid(), mesh=data.mesh, type=data.type, name=name, pmids=set()
                )
            node_dict_each.setdefault(name, self.node_dict[data.name])
            self.node_dict[name].pmids.add(data.pmid)
        else:
            # By MeSH
            # TODO: better way to deal with unseen MeSH types (e.g. Chromosome)
            if data.type == "Chromosome":
                return

            mesh_list = line.parse_mesh(self.mesh_map)

            for mesh in mesh_list:
                node_id = mesh["key"]
                if not self._node_id_registered(line, node_id):
                    self.node_dict[node_id] = PubTatorNodeData(
                        id=generate_uuid(),
                        mesh=mesh["mesh"],
                        type=data.type,
                        name=defaultdict(int),
                        pmids=set(),
                    )
                node_dict_each.setdefault(node_id, self.node_dict[node_id])
                self.node_dict[node_id].name[data.name] += 1
                self.node_dict[node_id].pmids.add(data.pmid)

    def _create_edges_for_relations(self, line: PubTatorLine):
        data = line.data
        # TODO: better way to deal with DNAMutation notation inconsistency
        mesh_1 = data.mesh1.split("|")[0]
        mesh_1 = self.mesh_map.get(mesh_1, mesh_1)
        mesh_2 = data.mesh2.split("|")[0]
        mesh_2 = self.mesh_map.get(mesh_2, mesh_2)
        self.edge_dict[(mesh_1, mesh_2)].append(
            PubTatorEdgeData(id=generate_uuid(), pmid=data.pmid, relation=data.relation)
        )

    def _node_id_registered(self, line: PubTatorLine, node_id: str):
        is_registered = False
        data = line.data
        if node_id in self.node_dict:
            is_registered = True
            if data.type != self.node_dict[node_id].type:
                info = {
                    "id": node_id,
                    "type": data.type,
                    "name": data.name,
                }
                logger.debug(f"Found collision of MeSH:\n{self.node_dict[node_id]}\n{info}")
                logger.debug("Discard the latter\n")

        return is_registered

    def _create_complete_graph(
        self,
        node_dict_each: dict[str, PubTatorNodeData],
        pmid: str,
    ):
        for i, name_1 in enumerate(node_dict_each.keys()):
            for j, name_2 in enumerate(node_dict_each.keys()):
                if i >= j:
                    continue
                if self.edge_dict.get((name_2, name_1), []):
                    key = (name_2, name_1)
                else:
                    key = (name_1, name_2)
                self.edge_dict[key].append(PubTatorEdgeData(id=generate_uuid(), pmid=pmid))

    def _determine_mesh_term_labels(self):
        def determine_node_label(candidate_name: str):
            nonlocal node_data
            return (node_data.name[candidate_name], -len(candidate_name))

        for node_id, node_data in self.node_dict.items():
            if isinstance(node_data.name, defaultdict):
                # Use the name that appears the most times
                self.node_dict[node_id].name = max(node_data.name, key=determine_node_label)

        self._merge_same_name_genes()

    def _merge_same_name_genes(self):
        gene_name_dict = defaultdict(list)
        for node_id, node_data in self.node_dict.items():
            if node_data.type != "Gene" or node_data.mesh in ("-", ""):
                continue
            gene_name_dict[node_data.name].append(
                {
                    "node_id": node_id,
                    "mesh": node_data.mesh,
                    "pmids": node_data.pmids,
                }
            )

        for node_data_list in gene_name_dict.values():
            mesh_list = []
            edges_to_merge = []
            removed_node_ids = []
            pmid_collection = set()
            for node_data in node_data_list:
                node_id = node_data["node_id"]
                # Nodes
                popped_node_data = self.node_dict.pop(node_id)
                removed_node_ids.append(node_id)
                mesh_list.append(node_data["mesh"])
                for pmid in node_data["pmids"]:
                    pmid_collection.add(pmid)

                # Edges
                for u, v in self.edge_dict.keys():
                    if u == node_id:
                        edges_to_merge.append((u, v, "u"))
                    elif v == node_id:
                        edges_to_merge.append((u, v, "v"))

            # Nodes
            concated_mesh = ";".join(mesh_list)
            # popped_node_data is the last node_data of nodes having the same name
            popped_node_data.mesh = concated_mesh
            popped_node_data.pmids = pmid_collection
            node_key = f"gene_{concated_mesh}"
            self.node_dict[node_key] = popped_node_data

            # Edges
            merged_edges = defaultdict(list)
            for u, v, pos in edges_to_merge:
                neighbor = v if pos == "u" else u
                try:
                    popped_edge_data = self.edge_dict.pop((u, v))
                except KeyError:
                    # The same gene but given different ids in the same article
                    assert neighbor in removed_node_ids
                if neighbor in removed_node_ids:
                    continue
                merged_edges[(node_key, neighbor)].extend(popped_edge_data)
            self.edge_dict.update(merged_edges)

    def _get_non_isolated_nodes(self):
        if self.node_type in ("all", "mesh"):
            self.non_isolated_nodes = set(self.node_dict.keys())
        elif self.node_type == "relation":
            for node_1, node_2 in self.edge_dict:
                self.non_isolated_nodes.add(node_1)
                self.non_isolated_nodes.add(node_2)

    def _parse_title(self, line: str):
        pmid, title = line.split("|t|")
        self.pmid_title_map.setdefault(pmid, title)
        return pmid

    @staticmethod
    def _is_title(line: str):
        return line.find("|t|") != -1

    @staticmethod
    def _parse_header(filepath: str):
        PubTatorLine.use_mesh = False
        with open(filepath) as f:
            for line in f.readlines():
                if not line.startswith(HEADER_SYMBOL):
                    break
                PubTatorParser._assign_flags(line.replace(HEADER_SYMBOL, "", 1).strip())

    @staticmethod
    def _assign_flags(line: str):
        if line == "USE-MESH-VOCABULARY":
            PubTatorLine.use_mesh = True



class EmptyInput(Exception):
    """Empty query or PMIDs"""

    def __init__(self, msg="Your search cannot be empty."):
        super().__init__(msg)


class NoArticles(Exception):
    """No articles found by PubTator3 API"""

    def __init__(self, msg="No articles found by PubTator3 API."):
        super().__init__(msg)


class UnsuccessfulRequest(Exception):
    """Unsuccessful request to PubTator3 API"""

    def __init__(self, msg="Unsuccessful request to PubTator3 API."):
        super().__init__(msg)



import os
from pathlib import Path
from tempfile import TemporaryDirectory
from types import SimpleNamespace
from uuid import uuid4

from dotenv import load_dotenv

load_dotenv()

TEMPDIR = TemporaryDirectory()
SAVEDIR = Path(TEMPDIR.name) if os.getenv("SAVEDIR") is None else Path(os.getenv("SAVEDIR"))
MAX_ARTICLES = 1000
DATA_FILENAME = {
    "graph": "G.pkl",
    "xgmml": "output.xgmml",
    "html": "output.html",
    "pubtator": "output.pubtator",
    "edge_info": "output.csv",
}
visibility = SimpleNamespace(visible={"visibility": "visible"}, hidden={"visibility": "hidden"})
display = SimpleNamespace(block={"display": "block"}, none={"display": "none"})


def generate_session_id():
    return str(uuid4())


def get_data_savepath(session_id: str):
    savepath = {}
    savedir = SAVEDIR / session_id
    savedir.mkdir(parents=True, exist_ok=True)
    for file, filepath in DATA_FILENAME.items():
        savepath[file] = str(savedir / filepath)
    return savepath


def clean_up_files():
    try:
        TEMPDIR.cleanup()
    except Exception:
        pass



import json
import pickle
from pathlib import Path
from queue import Queue
from types import SimpleNamespace
from unittest import mock

import pytest
import requests_mock

from netmedex.biocjson_parser import biocjson_to_pubtator, get_biocjson_annotations
from netmedex.exceptions import EmptyInput, NoArticles, UnsuccessfulRequest
from netmedex.pubtator_core import (
    PubTatorAPI,
    parse_cite_response,
    request_successful,
    send_publication_query,
    send_search_query,
    unsuccessful_query,
)
from netmedex.pubtator_utils import load_pmids

TESTDATA_DIR = Path(__file__).parent / "test_data"


@pytest.fixture(scope="module")
def paths():
    filepaths = {
        "json": TESTDATA_DIR / "pubtator3.37026113_240614.json",
        "json_full-text": TESTDATA_DIR / "pubtator3.22429397_full_240916.json",
        "pubtator-std_relation-std": TESTDATA_DIR
        / "pubtator3.37026113_standardized_std-relation_240614.pubtator",
        "pubtator-std_relation-ori": TESTDATA_DIR
        / "pubtator3.37026113_standardized_ori-relation_240614.pubtator",
        "pubtator": TESTDATA_DIR / "pubtator3.37026113_240614.pubtator",
        "tsv": TESTDATA_DIR / "cite_tsv.tsv",
        "pmids": TESTDATA_DIR / "pmid_list.txt",
        "query_search": TESTDATA_DIR / "search_N-dimethylnitrosamine_and_Metformin_241008.pkl",
        "query_cite": TESTDATA_DIR / "cite_N-dimethylnitrosamine_and_Metformin_241008.pkl",
    }
    return filepaths


@requests_mock.Mocker(kw="mock")
def test_send_search_query_search(paths, **kwargs):
    query = "N-dimethylnitrosamine and Metformin"

    with open(paths["query_search"], "rb") as f:
        res_json = pickle.load(f).json()

    kwargs["mock"].get(
        "https://www.ncbi.nlm.nih.gov/research/pubtator3-api/search/?text=N-dimethylnitrosamine+and+Metformin",
        status_code=200,
        json=res_json,
    )
    res = send_search_query(query, type="search")

    assert res.json()["count"] == 136


def test_api_send_search_query_search():
    query = "N-dimethylnitrosamine and Metformin"
    res = send_search_query(query, type="search")

    expected = 136
    if res.status_code == 200:
        total_articles = res.json()["count"]
    else:
        total_articles = None

    assert total_articles >= expected


@requests_mock.Mocker(kw="mock")
def test_send_search_query_cite(paths, **kwargs):
    query = "N-dimethylnitrosamine and Metformin"

    with open(paths["query_cite"], "rb") as f:
        res_text = pickle.load(f).text

    kwargs["mock"].get(
        "https://www.ncbi.nlm.nih.gov/research/pubtator3-api/cite/tsv?text=N-dimethylnitrosamine+and+Metformin",
        status_code=200,
        text=res_text,
    )

    res = send_search_query(query, type="cite")

    assert len(parse_cite_response(res.text)) == 136


def test_api_send_search_query_cite():
    query = "N-dimethylnitrosamine and Metformin"
    res = send_search_query(query, type="cite")
    expected = 130
    if res.status_code == 200:
        total_articles = len(parse_cite_response(res.text))
    else:
        total_articles = None

    assert total_articles >= expected


@requests_mock.Mocker(kw="mock")
def test_search_publication_abstract(paths, **kwargs):
    pmid = "37026113"

    with open(paths["json"]) as f:
        res_json = json.load(f)

    kwargs["mock"].get(
        "https://www.ncbi.nlm.nih.gov/research/pubtator3-api/publications/export/biocjson?pmids=37026113",
        status_code=200,
        json=res_json,
    )
    res = send_publication_query(pmid, article_id_type="pmids", format="biocjson", full_text=False)
    annotation_list = get_biocjson_annotations(res.json()["PubTator3"][0], retain_ori_text=False)

    assert len(annotation_list) == 33


@requests_mock.Mocker(kw="mock")
def test_search_publication_full_text(paths, **kwargs):
    pmid = "22429397"

    with open(paths["json_full-text"]) as f:
        res_json = json.load(f)

    kwargs["mock"].get(
        "https://www.ncbi.nlm.nih.gov/research/pubtator3-api/publications/export/biocjson?pmids=22429397&full=true",
        status_code=200,
        json=res_json,
    )

    res = send_publication_query(pmid, article_id_type="pmids", format="biocjson", full_text=True)
    annotation_list = get_biocjson_annotations(res.json()["PubTator3"][0], retain_ori_text=False)

    assert len(annotation_list) == 666


def test_api_search_publication_full_text():
    pmid = "37026113"
    res = send_publication_query(pmid, article_id_type="pmids", format="biocjson", full_text=True)

    expected_len = 379
    if res.status_code == 200:
        res = res.json()["PubTator3"][0]
        annotation_list = get_biocjson_annotations(res, retain_ori_text=False)
    else:
        annotation_list = []

    assert len(annotation_list) == expected_len


def test_biocjson_pubtator_equal(paths):
    pubtator_path = paths["pubtator"]
    biocjson_path = paths["json"]

    with open(pubtator_path) as pubtator, open(biocjson_path) as cjson:
        pubtator = pubtator.read()
        cjson = biocjson_to_pubtator(
            json.load(cjson), retain_ori_text=True, role_type="identifier"
        )
        assert cjson == pubtator


def test_parse_cite_response(paths):
    filepath = paths["tsv"]
    expected = ["38229736", "38173127", "38132876"]
    with open(filepath) as f:
        pmid_list = parse_cite_response(f.read())

    assert pmid_list == expected


def test_parse_biocjson_response(paths):
    filepath = paths["json"]
    with open(filepath) as f:
        content = json.load(f)["PubTator3"][0]

    expected_len = 33
    annotation_list = get_biocjson_annotations(content, retain_ori_text=False)

    assert len(annotation_list) == expected_len


@requests_mock.Mocker(kw="mock")
def test_use_mesh(paths, **kwargs):
    test_filepath = paths["pubtator-std_relation-std"]

    with open(paths["json"]) as f:
        res_json = json.load(f)

    kwargs["mock"].get(
        "https://www.ncbi.nlm.nih.gov/research/pubtator3-api/publications/export/biocjson?pmids=37026113",
        status_code=200,
        json=res_json,
    )

    pubtator_api = PubTatorAPI(
        query=None,
        pmid_list=["37026113"],
        savepath=None,
        search_type="pmids",
        sort="date",
        max_articles=1000,
        use_mesh=True,
        full_text=False,
        debug=False,
        queue=None,
    )
    output = pubtator_api.batch_publication_search()
    result = biocjson_to_pubtator(output[0], retain_ori_text=False)

    with open(test_filepath) as f:
        assert result == f.read()


@requests_mock.Mocker(kw="mock")
def test_batch_publication_queue(paths, **kwargs):
    kwargs["mock"].get(
        "https://www.ncbi.nlm.nih.gov/research/pubtator3-api/publications/export/pubtator",
        status_code=200,
        text="",
    )

    queue = Queue()
    progress = []
    expected = ["get/100/200", "get/200/200", None]
    pubtator_api = PubTatorAPI(
        query=None,
        pmid_list=[str(i) for i in range(200)],
        savepath=None,
        search_type="pmids",
        sort="date",
        max_articles=1000,
        use_mesh=False,
        full_text=False,
        debug=False,
        queue=queue,
    )
    pubtator_api.batch_publication_search()
    while True:
        result = queue.get()
        progress.append(result)
        if result is None:
            break

    assert progress == expected


@pytest.mark.parametrize(
    "args",
    [
        (None, None, "query"),
        ("   ", None, "query"),
        ([], None, "pmids"),
    ],
)
def test_empty_input(args):
    pubtator_api = PubTatorAPI(
        query=args[0],
        pmid_list=args[1],
        savepath=None,
        search_type=args[2],
        sort="date",
        max_articles=1000,
        use_mesh=False,
        full_text=False,
        debug=False,
        queue=None,
    )
    with pytest.raises(EmptyInput):
        pubtator_api.run()


@requests_mock.Mocker(kw="mock")
def test_no_articles(**kwargs):
    pubtator_api = PubTatorAPI(
        query="qwrsadga",
        pmid_list=None,
        savepath=None,
        search_type="query",
        sort="date",
        max_articles=1000,
        use_mesh=False,
        full_text=False,
        debug=False,
        queue=None,
    )
    kwargs["mock"].get(
        "https://www.ncbi.nlm.nih.gov/research/pubtator3-api/cite/tsv?text=qwrsadga",
        status_code=200,
        text="",
    )
    with pytest.raises(NoArticles):
        pubtator_api.run()


@requests_mock.Mocker(kw="mock")
@pytest.mark.parametrize(
    "type,full_text,use_mesh,file_format",
    [
        ("query", False, False, "pubtator"),
        ("query", True, False, "biocjson"),
        ("query", False, True, "biocjson"),
        ("query", True, True, "biocjson"),
    ],
)
def test_run_query_pipeline_query(type, full_text, use_mesh, file_format, paths, **kwargs):
    with open(paths["query_cite"], "rb") as f:
        res_text = pickle.load(f).text

    kwargs["mock"].get(
        "https://www.ncbi.nlm.nih.gov/research/pubtator3-api/cite/tsv?text=N-dimethylnitrosamine+and+Metformin",
        status_code=200,
        text=res_text,
    )

    text_or_json = {}
    if file_format == "pubtator":
        text_or_json["text"] = "foobar"
    elif file_format == "biocjson":
        text_or_json["json"] = {"PubTator3": []}

    kwargs["mock"].get(
        f"https://www.ncbi.nlm.nih.gov/research/pubtator3-api/publications/export/{file_format}",
        status_code=200,
        **text_or_json,
    )
    pubtator_api = PubTatorAPI(
        query="N-dimethylnitrosamine and Metformin",
        pmid_list=None,
        savepath=None,
        search_type=type,
        sort="date",
        max_articles=1000,
        use_mesh=use_mesh,
        full_text=full_text,
        debug=False,
        queue=None,
    )
    pubtator_api.run()


def test_load_pmids_file(paths):
    pmid_list = load_pmids(paths["pmids"], load_from="file")

    assert pmid_list == ["34205807", "34895069", "35883435"]


@pytest.mark.parametrize(
    "query,expected",
    [
        ("34205807, 34895069, 35883435", ["34205807", "34895069", "35883435"]),
        ("34205807,34895069,35883435", ["34205807", "34895069", "35883435"]),
        ("foo,bar,foobar", []),
        ("", []),
        (None, []),
    ],
)
def test_load_pmids_string(query, expected):
    pmid_list = load_pmids(query, load_from="string")

    assert pmid_list == expected


@pytest.mark.parametrize(
    "status_code,expected",
    [
        (200, True),
        (404, False),
        (502, False),
    ],
)
def test_request_successful(status_code, expected, **kwargs):
    res = SimpleNamespace(status_code=status_code)
    with mock.patch("netmedex.pubtator_core.logger") as mock_logger:
        assert request_successful(res) == expected
        if not expected:
            mock_logger.info.assert_called_with("Unsuccessful request")
            mock_logger.debug.assert_called_with(f"Response status code: {status_code}")


@pytest.mark.parametrize(
    "status_code,msg",
    [
        (404, "Please retry later."),
        (502, "Possibly too many articles. Please try more specific queries."),
    ],
)
def test_unsuccessful_query(status_code, msg, **kwargs):
    with pytest.raises(UnsuccessfulRequest):
        with mock.patch("netmedex.pubtator_core.logger") as mock_logger:
            unsuccessful_query(status_code)
            mock_logger.warning.assert_called_with(msg)



import base64
import pickle
import threading
from queue import Queue

from dash import Input, Output, State, no_update

from netmedex.exceptions import EmptyInput, NoArticles, UnsuccessfulRequest
from netmedex.network_core import NetworkBuilder
from netmedex.pubtator_core import PubTatorAPI
from netmedex.pubtator_utils import load_pmids
from netmedex.utils_threading import run_thread_with_error_notification
from webapp.utils import generate_session_id, get_data_savepath, visibility


def callbacks(app):
    @app.long_callback(
        Output("cy-graph-container", "style", allow_duplicate=True),
        Output("memory-graph-cut-weight", "data", allow_duplicate=True),
        Output("is-new-graph", "data"),
        Output("pmid-title-dict", "data"),
        Output("current-session-path", "data"),
        Input("submit-button", "n_clicks"),
        [
            State("api-toggle-items", "value"),
            State("sort-toggle-methods", "value"),
            State("input-type-selection", "value"),
            State("data-input", "value"),
            State("pmid-file-data", "contents"),
            State("pubtator-file-data", "contents"),
            State("cut-weight", "value"),
            State("max-edges", "value"),
            State("max-articles", "value"),
            State("pubtator-params", "value"),
            State("cy-params", "value"),
            State("weighting-method", "value"),
            State("node-type", "value"),
        ],
        running=[(Input("submit-button", "disabled"), True, False)],
        progress=[
            Output("progress", "value"),
            Output("progress", "max"),
            Output("progress", "label"),
            Output("progress-status", "children"),
        ],
        prevent_initial_call=True,
    )
    def run_pubtator3_api(
        set_progress,
        btn,
        source,
        sort_by,
        input_type,
        data_input,
        pmid_file_data,
        pubtator_file_data,
        weight,
        max_edges,
        max_articles,
        pubtator_params,
        cy_params,
        weighting_method,
        node_type,
    ):
        _exception_msg = None
        _exception_type = None

        def custom_hook(args):
            nonlocal _exception_msg
            nonlocal _exception_type
            _exception_msg = args.exc_value
            _exception_type = args.exc_type

        use_mesh = "use_mesh" in pubtator_params
        full_text = "full_text" in pubtator_params
        community = "community" in cy_params
        savepath = get_data_savepath(generate_session_id())

        query = None
        pmid_list = None
        if source == "api":
            if input_type == "query":
                query = data_input
            elif input_type == "pmids":
                pmid_list = load_pmids(data_input, load_from="string")
            elif input_type == "pmid_file":
                content_type, content_string = pmid_file_data.split(",")
                decoded_content = base64.b64decode(content_string).decode("utf-8")
                decoded_content = decoded_content.replace("\n", ",")
                pmid_list = load_pmids(decoded_content, load_from="string")
                input_type = "pmids"

            queue = Queue()
            threading.excepthook = custom_hook
            pubtator_api = PubTatorAPI(
                query=query,
                pmid_list=pmid_list,
                savepath=savepath["pubtator"],
                search_type=input_type,
                sort=sort_by,
                max_articles=max_articles,
                full_text=full_text,
                use_mesh=use_mesh,
                debug=False,
                queue=queue,
            )
            job = threading.Thread(
                target=run_thread_with_error_notification(pubtator_api.run, queue),
            )
            set_progress((0, 1, "", "(Step 1/2) Finding articles..."))

            job.start()
            while True:
                progress = queue.get()
                if progress is None:
                    break
                status, n, total = progress.split("/")
                if status.startswith("search"):
                    status_msg = "(Step 1/2) Finding articles..."
                elif status == "get":
                    status_msg = "(Step 2/2) Retrieving articles..."
                progress_bar_msg = f"{n}/{total}"
                set_progress((int(n), int(total), progress_bar_msg, status_msg))

            if _exception_type is not None:
                known_exceptions = (
                    EmptyInput,
                    NoArticles,
                    UnsuccessfulRequest,
                )
                if issubclass(_exception_type, known_exceptions):
                    exception_msg = str(_exception_msg)
                else:
                    exception_msg = "An unexpected error occurred."
                set_progress((1, 1, "", exception_msg))
                return (no_update, weight, False, no_update, no_update)

            job.join()
        elif source == "file":
            with open(savepath["pubtator"], "w") as f:
                content_type, content_string = pubtator_file_data.split(",")
                decoded_content = base64.b64decode(content_string).decode("utf-8")
                f.write(decoded_content)

        set_progress((0, 1, "0/1", "Generating network..."))
        G = NetworkBuilder(
            pubtator_filepath=savepath["pubtator"],
            savepath=savepath["html"],
            node_type=node_type,
            output_filetype="html",
            weighting_method=weighting_method,
            edge_weight_cutoff=0,
            pmid_weight_filepath=None,
            community=False,
            max_edges=0,
            debug=False,
        ).run()

        # Keeping track of the graph's metadata
        G.graph["is_community"] = True if community else False
        G.graph["max_edges"] = max_edges

        with open(savepath["graph"], "wb") as f:
            pickle.dump(G, f)

        return (visibility.visible, weight, True, G.graph["pmid_title"], savepath)



import dash
from dash import html


def generate_param_title(title, descriptions, is_right=False):
    class_name = "info-outer info-right" if is_right else "info-outer info-left"
    kwargs = {"data-tooltip": descriptions, "data-x": "0px", "data-y": "0px"}
    return html.Div(
        [
            html.H5(title),
            html.Span(
                [
                    html.Img(src=dash.get_asset_url("icon_info.svg"), className="info-img"),
                ],
                className=class_name,
                **kwargs,
            ),
        ],
        className="param-title",
    )


def icon_download():
    return html.Img(
        src=dash.get_asset_url("icon_download.svg"),
        width=20,
        height=20,
        style={"margin-right": "5px"},
    )



IRREGULAR_WORDS = {"species": "species"}


def s_stemmer(word: str):
    subwords = word.split(" ")
    last_subword = subwords[-1]

    if last_subword in IRREGULAR_WORDS:
        return " ".join(subwords[:-1] + [IRREGULAR_WORDS[last_subword]])

    if last_subword.endswith("ies") and not last_subword.endswith(("eies", "aies")):
        word = word[:-3] + "y"
    elif last_subword.endswith("es") and not last_subword.endswith(("aes", "ees", "oes")):
        word = word[:-1]
    elif last_subword.endswith("s") and not last_subword.endswith(("is", "us", "ss")):
        word = word[:-1]

    return word



from dash import Input, Output, State


def callbacks(app):
    @app.callback(
        Output("advanced-settings-collapse", "style", allow_duplicate=True),
        Output("max-edges", "tooltip"),
        Output("max-articles", "tooltip"),
        Input("advanced-settings-btn", "n_clicks"),
        State("advanced-settings-collapse", "style"),
        prevent_initial_call=True,
    )
    def open_advanced_options(n_clicks, style):
        visibility = style["visibility"]
        toggle = {"hidden": "visible", "visible": "hidden"}
        weight_toggle = {"hidden": True, "visible": False}
        return (
            {"visibility": toggle[visibility]},
            {"placement": "bottom", "always_visible": weight_toggle[visibility]},
            {"placement": "bottom", "always_visible": weight_toggle[visibility]},
        )



import pytest

from netmedex.stemmers import s_stemmer


@pytest.mark.parametrize(
    "input,expected",
    [
        ("", ""),
        ("abdominal diseases", "abdominal disease"),
        ("reactive oxygen species", "reactive oxygen species"),
    ],
)
def test_s_stemmer(input, expected):
    output = s_stemmer(input)
    assert output == expected, f"result: {output}, expected: {expected}"



import json
from pathlib import Path

import pytest

from netmedex.biocjson_parser import biocjson_to_pubtator


@pytest.fixture(scope="module")
def paths(request):
    test_dir = Path(request.config.rootdir) / "tests/test_data"
    paths = {
        "abstract_json": test_dir / "pubtator3.22439397_abstract_240916.json",
        "full_json": test_dir / "pubtator3.22429397_full_240916.json",
        "abstract_pubtator": test_dir / "22429397_abstract_240916.pubtator",
        "full_pubtator": test_dir / "22429397_full_240916.pubtator",
        "full_to_abstract_pubtator": test_dir / "22429397_full_to_abstract_240916.pubtator",
    }
    return paths


def test_abstract_to_abstract_conversion(paths):
    with open(paths["abstract_json"]) as f:
        data = json.load(f)

    result = biocjson_to_pubtator(data, retain_ori_text=True, only_abstract=True)

    with open(paths["abstract_pubtator"]) as f:
        expected = f.read()

    assert result == expected


def test_full_to_abstract_conversion(paths):
    with open(paths["full_json"]) as f:
        data = json.load(f)

    result = biocjson_to_pubtator(data, retain_ori_text=True, only_abstract=True)

    with open(paths["full_to_abstract_pubtator"]) as f:
        expected = f.read()

    assert result == expected


def test_full_conversion(paths):
    with open(paths["full_json"]) as f:
        data = json.load(f)

    result = biocjson_to_pubtator(data, retain_ori_text=True, only_abstract=False)

    with open(paths["full_pubtator"]) as f:
        expected = f.read()

    assert result == expected



from queue import Queue


def run_thread_with_error_notification(func, queue: Queue):
    """Add None to the queue if any exception occurs."""

    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception as e:
            queue.put(None)
            raise e

    return wrapper



import logging
from collections import defaultdict
from typing import Literal

logger = logging.getLogger(__name__)


def biocjson_to_pubtator(
    res_json,
    retain_ori_text: bool = True,
    only_abstract: bool = False,
    role_type: Literal["identifier", "name"] = "identifier",
):
    # 2024/05/26: PubTator has changed the format of the response
    res_json = res_json["PubTator3"]

    converted_strs = []
    for each_res_json in res_json:
        pmid = each_res_json["pmid"]

        title_passage = extract_passage(each_res_json, "TITLE")
        abstract_passage = extract_passage(each_res_json, "ABSTRACT")

        if not title_passage or not abstract_passage:
            continue

        if only_abstract:
            abstract_idx = title_passage["idx"] + abstract_passage["idx"]
        else:
            abstract_idx = None

        annotation_list = get_biocjson_annotations(
            each_res_json, retain_ori_text, abstract_idx=abstract_idx
        )
        relation_list = get_biocjson_relations(each_res_json, role_type)
        converted_strs.append(
            create_pubtator_str(
                pmid,
                # Only one title passage
                title_passage["text"][0],
                " ".join(abstract_passage["text"]),
                annotation_list,
                relation_list,
            )
        )

    return "".join(converted_strs)


def extract_passage(content, name):
    passage_info = defaultdict(list)
    # "section_type" exists if the article has full text
    try:
        content["passages"][0]["infons"]["section_type"]
        section_type = "section_type"
    except KeyError:
        section_type = "type"
        name = name.lower()

    for idx, passage in enumerate(content["passages"]):
        if passage["infons"][section_type] == name:
            passage_json = content["passages"][idx]
            passage_idx = idx
            passage_info["text"].append(passage_json["text"])
            passage_info["idx"].append(passage_idx)

    return passage_info


def get_biocjson_annotations(res_json, retain_ori_text, abstract_idx=None):
    n_passages = len(res_json["passages"])

    annotation_list = []
    # TODO: extract from specific passages only (if full_text)?
    if abstract_idx:
        passages = [res_json["passages"][i]["annotations"] for i in abstract_idx]
    else:
        passages = [res_json["passages"][i]["annotations"] for i in range(n_passages)]
    for annotation_entries in passages:
        for annotation_entry in annotation_entries:
            annotation = {}
            try:
                id = annotation_entry["infons"]["identifier"]
            except Exception:
                id = "-"
            annotation["id"] = "-" if id == "None" or id is None else id
            annotation["type"] = annotation_entry["infons"]["type"]
            annotation["locations"] = annotation_entry["locations"][0]
            annotation["name"] = get_name(retain_ori_text, annotation_entry, annotation)
            if annotation["type"] == "Variant":
                annotation["type"] = annotation_entry["infons"]["subtype"]

            if annotation["name"] is None:
                continue
            annotation_list.append(annotation)

    return annotation_list


def get_name(retain_ori_text, annotation_entry, annotation):
    try:
        if retain_ori_text:
            name = annotation_entry["text"]
            # In type == "species", the entity name is stored in "text"
        elif annotation["type"] == "Species":
            name = annotation_entry["text"]
            # Variant can be either SNP, DNAMutation, or ProteinMutation
        elif annotation["type"] == "Variant":
            # Some variants may not have standardized name
            try:
                name = annotation_entry["infons"]["name"]
            except KeyError:
                name = None
        elif annotation_entry["infons"].get("database", "none") == "omim":
            name = annotation_entry["text"]
        else:
            try:
                name = annotation_entry["infons"]["name"]
            except KeyError:
                name = annotation_entry["text"]
    except KeyError as e:
        name = None
        logger.warning(f"Cannot find annotation name: {str(e)}")

    return name


def get_biocjson_relations(res_json, role_type):
    relation_list = []
    for relation_entry in res_json["relations"]:
        each_relation = {}
        each_relation["role1"] = relation_entry["infons"]["role1"][role_type]
        each_relation["role2"] = relation_entry["infons"]["role2"][role_type]
        each_relation["type"] = relation_entry["infons"]["type"]
        relation_list.append(each_relation)

    return relation_list


def create_pubtator_str(pmid, title, abstract, annotation_list, relation_list):
    title_str = f"{pmid}|t|{title}\n"
    abstract_str = f"{pmid}|a|{abstract}\n"
    annotation_list.sort(key=lambda x: x["locations"]["offset"])
    annotation_str = [
        (
            f"{pmid}\t"
            f"{annotation['locations']['offset']}\t"
            f"{annotation['locations']['length'] + annotation['locations']['offset']}\t"
            f"{annotation['name']}\t"
            f"{annotation['type']}\t"
            f"{annotation['id']}"
        )
        for annotation in annotation_list
    ]
    relation_str = [
        (f"{pmid}\t" f"{relation['type']}\t" f"{relation['role1']}\t" f"{relation['role2']}")
        for relation in relation_list
    ]

    return (
        title_str
        + abstract_str
        + "\n".join(annotation_str)
        + "\n"
        + "\n".join(relation_str)
        + "\n\n"
    )



import dash
import dash_bootstrap_components as dbc
from dash import dcc, html

from webapp.components.utils import generate_param_title, icon_download
from webapp.utils import visibility

graph_layout = html.Div(
    [
        generate_param_title(
            "Graph Layout",
            "Select a layout to arrange the nodes",
            is_right=True,
        ),
        dcc.Dropdown(
            id="graph-layout",
            options=[
                {"label": "Preset", "value": "preset"},
                {"label": "Circle", "value": "circle"},
                {"label": "Grid", "value": "grid"},
                {"label": "Random", "value": "random"},
                {"label": "Concentric", "value": "concentric"},
                {"label": "Breadthfirst", "value": "breadthfirst"},
                {"label": "Cose", "value": "cose"},
            ],
            value="preset",
            style={"width": "200px"},
        ),
    ],
    className="param",
)


minimal_degree = html.Div(
    [
        generate_param_title(
            "Minimal Degree",
            "Set the minimum node degree to filter the graph",
            is_right=True,
        ),
        dbc.Input(
            id="node-degree",
            min=1,
            step=1,
            value=1,
            type="number",
            style={"width": "200px"},
        ),
        dcc.Store(id="memory-node-degree", data=1),
    ],
    className="param",
)

edge_weight_cutoff = html.Div(
    [
        generate_param_title(
            "Edge Weight Cutoff",
            "Set the minimum edge weight to filter the graph",
            is_right=True,
        ),
        dcc.Slider(
            0,
            20,
            1,
            value=3,
            marks=None,
            id="graph-cut-weight",
            tooltip={"placement": "bottom", "always_visible": False},
        ),
        dcc.Store(id="memory-graph-cut-weight", data=3),
    ],
    className="param",
)

graph_tools = html.Div(
    [
        dbc.Button(
            [
                icon_download(),
                "PubTator",
            ],
            id="download-pubtator-btn",
            className="export-btn",
            color="success",
            style=visibility.hidden,
        ),
        dcc.Download(id="download-pubtator"),
        dbc.Button(
            [
                icon_download(),
                "HTML",
            ],
            id="export-btn-html",
            className="export-btn",
        ),
        dcc.Download(id="export-html"),
        dbc.Button(
            [
                icon_download(),
                "XGMML",
            ],
            id="export-btn-xgmml",
            className="export-btn",
        ),
        dcc.Download(id="export-xgmml"),
        dbc.Button(
            html.Img(
                src=dash.get_asset_url("three-dots.svg"),
            ),
            # "Settings",
            id="graph-settings-btn",
            className="btn-secondary settings",
        ),
        html.Div(
            [
                html.H4("Settings"),
                graph_layout,
                edge_weight_cutoff,
                minimal_degree,
            ],
            id="graph-settings-collapse",
            className="settings-collapse",
            style=visibility.hidden,
        ),
    ],
    className="settings-container settings-background",
)



import logging
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

import pytest

from netmedex.cli import main
from netmedex.network_core import NetworkBuilder
from netmedex.pubtator_core import PubTatorAPI

logging.basicConfig(level=logging.DEBUG)


@pytest.fixture(scope="module")
def tempdir():
    with TemporaryDirectory() as tempdir:
        yield Path(tempdir)


@pytest.fixture(scope="module")
def paths(request):
    test_dir = Path(request.config.rootdir) / "tests/test_data"
    return {"simple": test_dir / "6_nodes_3_clusters_mesh.pubtator"}


@pytest.fixture(scope="module")
def network_cli_args(paths):
    args = {
        "pubtator_filepath": paths["simple"],
        "savepath": None,
        "node_type": "all",
        "output_filetype": "html",
        "weighting_method": "freq",
        "edge_weight_cutoff": 1,
        "pmid_weight_filepath": None,
        "max_edges": 0,
        "community": False,
        "debug": False,
    }
    return {
        "args_basic": args,
        "args_community": {**args, "community": True},
        "args_mesh_only": {**args, "node_type": "mesh"},
        "args_relation_only": {**args, "node_type": "relation"},
        "args_npmi": {**args, "weighting_method": "npmi"},
        "args_weight": {**args, "edge_weight_cutoff": 20},
        "args_xgmml": {**args, "output_filetype": "xgmml"},
    }


@pytest.mark.parametrize(
    "query,error_msg",
    [
        ("", "Your search cannot be empty."),
        ("qtihasiogha", "No articles found by PubTator3 API."),
        ("covid19", "Possibly too many articles. Please try more specific queries."),
    ],
)
def test_api_pubtator3_api_error(query, error_msg, tempdir, monkeypatch: pytest.MonkeyPatch):
    args = ["netmedex", "search", "-q", query, "-o", str(tempdir / "test.pubtator")]
    monkeypatch.setattr("sys.argv", args)
    with (
        mock.patch("netmedex.pubtator_core.FALLBACK_SEARCH", False),
        mock.patch("netmedex.cli.logger") as mock_logger,
    ):
        main()
        mock_logger.error.assert_called_with(error_msg)


def test_api_pubtator3_api_fallback_search(tempdir, monkeypatch: pytest.MonkeyPatch):
    args = [
        "netmedex",
        "search",
        "-q",
        "covid19",
        "-o",
        str(tempdir / "test.pubtator"),
        "--max_articles",
        "10",
    ]
    monkeypatch.setattr("sys.argv", args)
    main()


@pytest.mark.parametrize(
    "args,expected",
    [
        (
            [
                "netmedex",
                "search",
                "-q",
                "foo",
                "-o",
                "bar",
                "--max_articles",
                "100",
                "--full_text",
                "-s",
                "date",
            ],
            {
                "query": "foo",
                "pmid_list": None,
                "savepath": "bar",
                "search_type": "query",
                "sort": "date",
                "max_articles": 100,
                "full_text": True,
                "use_mesh": False,
                "debug": False,
                "queue": None,
            },
        ),
        (
            [
                "netmedex",
                "search",
                "-q",
                "foo",
                "-o",
                "bar",
                "--max_articles",
                "100",
                "--use_mesh",
                "-s",
                "score",
            ],
            {
                "query": "foo",
                "pmid_list": None,
                "savepath": "bar",
                "search_type": "query",
                "sort": "score",
                "max_articles": 100,
                "full_text": False,
                "use_mesh": True,
                "debug": False,
                "queue": None,
            },
        ),
        (
            ["netmedex", "search", "-q", "foo"],
            {
                "query": "foo",
                "pmid_list": None,
                "savepath": "query_foo.pubtator",
                "search_type": "query",
                "sort": "date",
                "max_articles": 1000,
                "full_text": False,
                "use_mesh": False,
                "debug": False,
                "queue": None,
            },
        ),
        (
            ["netmedex", "search", "-p", "123,456", "-o", "bar"],
            {
                "query": None,
                "pmid_list": ["123", "456"],
                "savepath": "bar",
                "search_type": "pmids",
                "sort": "date",
                "max_articles": 1000,
                "full_text": False,
                "use_mesh": False,
                "debug": False,
                "queue": None,
            },
        ),
        (
            ["netmedex", "search", "-p", "123,456", "--full_text"],
            {
                "query": None,
                "pmid_list": ["123", "456"],
                "savepath": "pmids_123_total_2.pubtator",
                "search_type": "pmids",
                "sort": "date",
                "max_articles": 1000,
                "full_text": True,
                "use_mesh": False,
                "debug": False,
                "queue": None,
            },
        ),
        (
            [
                "netmedex",
                "search",
                "-f",
                "./tests/test_data/pmid_list.txt",
                "--max_articles",
                "10000",
            ],
            {
                "query": None,
                "pmid_list": ["34205807", "34895069", "35883435"],
                "savepath": "pmids_34205807_total_3.pubtator",
                "search_type": "pmids",
                "sort": "date",
                "max_articles": 10000,
                "full_text": False,
                "use_mesh": False,
                "debug": False,
                "queue": None,
            },
        ),
    ],
)
def test_pubtator_api_main(args, expected, monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("sys.argv", args)
    with mock.patch("netmedex.pubtator_core.PubTatorAPI") as mock_pipeline:
        main()
        mock_pipeline.assert_called_once_with(
            query=expected["query"],
            pmid_list=expected["pmid_list"],
            savepath=expected["savepath"],
            search_type=expected["search_type"],
            sort=expected["sort"],
            max_articles=expected["max_articles"],
            use_mesh=expected["use_mesh"],
            full_text=expected["full_text"],
            debug=expected["debug"],
            queue=expected["queue"],
        )


@pytest.mark.parametrize(
    "args",
    [
        ["netmedex", "search", "-q", ""],
        ["netmedex", "search", "-p", "   "],
    ],
)
def test_pubtator3_api_exceptions(args, monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("sys.argv", args)
    with mock.patch("netmedex.cli.logger") as mock_logger:
        main()
        mock_logger.error.assert_called_once()


@pytest.mark.parametrize(
    "args",
    [
        ["netmedex", "search", "-p", "123", "-q", "bar"],
        ["netmedex", "search", "-f", "foo.txt", "-q", "bar"],
    ],
)
def test_pubtator3_api_exit(args, monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("sys.argv", args)
    with pytest.raises(SystemExit):
        with mock.patch("netmedex.cli.logger") as mock_logger:
            main()
            mock_logger.info.assert_called_once()


def test_write_output_none():
    open_mock = mock.mock_open()
    args = {
        "query": None,
        "pmid_list": ["123", "456"],
        "savepath": None,
        "search_type": "pmids",
        "sort": "date",
        "max_articles": 1000,
        "full_text": True,
        "use_mesh": False,
        "debug": False,
        "queue": None,
    }
    with mock.patch("builtins.open", create=True):
        pubtator3_api = PubTatorAPI(**args)
        pubtator3_api._write_results("foo")
        open_mock.assert_not_called()


@pytest.mark.parametrize(
    "output,savepath,use_mesh,expected",
    [("foo", "bar.pubtator", False, "foo"), ("foo", "bar.pubtator", True, "foo")],
)
def test_write_output(output, savepath, use_mesh, expected):
    open_mock = mock.mock_open()
    args = {
        "query": None,
        "pmid_list": ["123", "456"],
        "savepath": savepath,
        "search_type": "pmids",
        "sort": "date",
        "max_articles": 1000,
        "full_text": True,
        "use_mesh": use_mesh,
        "debug": False,
        "queue": None,
    }
    with (
        mock.patch("builtins.open", open_mock, create=True),
        mock.patch("netmedex.pubtator_core.logger") as mock_logger,
    ):
        pubtator3_api = PubTatorAPI(**args)
        pubtator3_api._write_results(output)
        if use_mesh:
            open_mock.return_value.writelines.assert_any_call(["##USE-MESH-VOCABULARY", "\n"])
        open_mock.return_value.writelines.assert_any_call(output)
        mock_logger.info.assert_called_with(f"Save to {savepath}")


@pytest.mark.parametrize(
    "args",
    [
        "args_basic",
        "args_community",
        "args_mesh_only",
        "args_relation_only",
        "args_npmi",
        "args_weight",
        "args_xgmml",
    ],
)
def test_network_cli(args, paths, tempdir, network_cli_args):
    network_builder = NetworkBuilder(**network_cli_args[args])
    network_builder.run()



from collections.abc import Sequence
from dataclasses import dataclass


@dataclass
class GraphNodeData:
    _id: str
    color: str
    label_color: str
    shape: str
    type: str
    name: str
    document_frequency: int
    marked: bool = False
    parent: str | None = None
    pos: Sequence[float] | None = None


@dataclass
class GraphEdgeData:
    _id: str
    raw_frequency: int
    doc_weighted_frequency: float
    npmi: float
    edge_weight: float
    pmids: list[str]
    scaled_edge_weight: float | None = None
    edge_width: int | None = None


@dataclass
class CommunityEdgeData:
    _id: int
    edge_weight: float
    scaled_edge_weight: float
    edge_width: int
    pmids: list[str]



import dash
from dash import dcc, html

from webapp.components.graph_info import graph_info
from webapp.components.graph_tools import graph_tools
from webapp.utils import visibility

graph = html.Div(
    [
        html.Div(
            [
                html.Img(src=dash.get_asset_url("NetMedEx.png"), height="40px"),
            ],
            className="d-flex flex-row justify-content-center",
        ),
        html.Div(
            [
                graph_tools,
                graph_info,
                html.Div(id="cy-graph", className="flex-grow-1"),
                dcc.Store(id="is-new-graph", data=False),
                dcc.Store(id="pmid-title-dict", data={}),
            ],
            id="cy-graph-container",
            className="d-flex flex-column flex-grow-1 position-relative",
            style=visibility.hidden,
        ),
    ],
    className="d-flex flex-column flex-grow-1 main-div",
)



from pathlib import Path

import networkx as nx
from lxml import etree
from lxml.builder import E
from lxml.etree import QName

XML_NAMESPACE = {
    "cy": "http://www.cytoscape.org",
    "dc": "http://purl.org/dc/elements/1.1/",
    "xlink": "http://www.w3.org/1999/xlink",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
}


for prefix, uri in XML_NAMESPACE.items():
    etree.register_namespace(prefix, uri)


TYPE_ATTR = {
    "string": {"type": "string", QName(XML_NAMESPACE["cy"], "type"): "String"},
    "boolean": {"type": "boolean", QName(XML_NAMESPACE["cy"], "type"): "Boolean"},
    "integer": {"type": "integer", QName(XML_NAMESPACE["cy"], "type"): "Integer"},
    "double": {"type": "double", QName(XML_NAMESPACE["cy"], "type"): "Double"},
}


def save_as_xgmml(G: nx.Graph, savepath):
    with open(savepath, "wb") as f:
        graph = create_graph_xml(G, Path(savepath).stem)
        f.write(
            etree.tostring(
                graph, encoding="utf-8", xml_declaration=True, standalone="yes", pretty_print=True
            )
        )


def create_graph_xml(G, graph_label="0"):
    _dummy_attr = {
        QName(XML_NAMESPACE["dc"], "dummy"): "",
        QName(XML_NAMESPACE["xlink"], "dummy"): "",
        QName(XML_NAMESPACE["rdf"], "dummy"): "",
    }
    _graph_attr = {
        "id": "0",
        "label": graph_label,
        "directed": "1",
        "xmlns": "http://www.cs.rpi.edu/XGMML",
        QName(XML_NAMESPACE["cy"], "documentVersion"): "3.0",
        **_dummy_attr,
    }

    graph = E.graph(_graph_attr, create_graphic_xml(), *create_node_xml(G), *create_edge_xml(G))

    # Delete attributes, keep namespace definition only
    for key in _dummy_attr:
        del graph.attrib[key]

    return graph


def create_graphic_xml():
    graph = E.graphics(
        # E.att(name_value("NETWORK_WIDTH", "795.0")),
        # E.att(name_value("NETWORK_DEPTH", "0.0")),
        # E.att(name_value("NETWORK_HEIGHT", "500.0")),
        # E.att(name_value("NETWORK_NODE_SELECTION", "true")),
        # E.att(name_value("NETWORK_EDGE_SELECTION", "true")),
        # E.att(name_value("NETWORK_BACKGROUND_PAINT", "#FFFFFF")),
        # E.att(name_value("NETWORK_CENTER_Z_LOCATION", "0.0")),
        # E.att(name_value("NETWORK_NODE_LABEL_SELECTION", "false")),
        # E.att(name_value("NETWORK_TITLE", "")),
    )

    return graph


def create_node_xml(G):
    node_collection = []
    for node in G.nodes(data=True):
        node_collection.append(_create_node_xml(node))

    return node_collection


def _create_node_xml(node):
    node_id, node_attr = node

    _node_attr = {"id": node_attr["_id"], "label": node_attr["name"]}
    _graphics_attr = {
        "width": "0.0",
        "h": "35.0",
        "w": "35.0",
        "z": "0.0",
        "x": str(round(node_attr["pos"][0], 3)),
        "y": str(round(node_attr["pos"][1], 3)),
        "type": node_attr["shape"],
        "outline": "#CCCCCC",
        "fill": node_attr["color"],
    }
    if node_attr["marked"]:
        _graphics_attr["outline"] = "#CF382C"
        _graphics_attr["width"] = "5.0"

    node = E.node(
        _node_attr,
        E.att(name_value("shared name", node_attr["name"])),
        E.att(name_value("name", node_attr["name"])),
        E.att(name_value("class", node_attr["type"])),
        E.graphics(
            _graphics_attr,
            # E.att(name_value("NODE_SELECTED", "false")),
            # E.att(name_value("NODE_NESTED_NETWORK_IMAGE_VISIBLE", "true")),
            # E.att(name_value("NODE_DEPTH", "0.0")),
            # E.att(name_value("NODE_SELECTED_PAINT", "#FFFF00")),
            # E.att(name_value("NODE_LABEL_ROTATION", "0.0")),
            # E.att(name_value("NODE_LABEL_WIDTH", "200.0")),
            # E.att(name_value("COMPOUND_NODE_PADDING", "10.0")),
            # E.att(name_value("NODE_LABEL_TRANSPARENCY", "255")),
            # E.att(name_value("NODE_LABEL_POSITION", "C,C,c,0.00,0.00")),
            # E.att(name_value("NODE_LABEL", node_attr["name"])),
            # E.att(name_value("NODE_VISIBLE", "true")),
            # E.att(name_value("NODE_LABEL_FONT_SIZE", "12")),
            # E.att(name_value("NODE_BORDER_STROKE", "SOLID")),
            # E.att(name_value("NODE_LABEL_FONT_FACE",
            #                  "SansSerif.plain,plain,12")),
            # E.att(name_value("NODE_BORDER_TRANSPARENCY", "255")),
            # E.att(name_value("COMPOUND_NODE_SHAPE", node_attr["shape"])),
            # E.att(name_value("NODE_LABEL_COLOR", "#000000")),
            # E.att(name_value("NODE_TRANSPARENCY", "255")),
        ),
    )

    return node


def name_value(name, value, with_type="string"):
    attr = {"name": name, "value": value}
    if with_type is not None:
        attr.update(TYPE_ATTR[with_type])
    return attr


def create_edge_xml(G):
    edge_collection = []
    for edge in G.edges(data=True):
        edge_collection.append(_create_edge_xml(edge, G))

    return edge_collection


def _create_edge_xml(edge, G):
    node_id_1, node_id_2, edge_attr = edge

    _edge_attr = {
        "id": edge_attr["_id"],
        "label": f"{G.nodes[node_id_1]['name']} (interacts with) {G.nodes[node_id_2]['name']}",
        "source": G.nodes[node_id_1]["_id"],
        "target": G.nodes[node_id_2]["_id"],
        QName(XML_NAMESPACE["cy"], "directed"): "1",
    }

    _graphics_attr = {"width": str(edge_attr["scaled_edge_weight"]), "fill": "#848484"}

    edge = E.edge(
        _edge_attr,
        E.att(
            name_value(
                "shared name",
                f"{G.nodes[node_id_1]['name']} (interacts with) {G.nodes[node_id_2]['name']}",
            )
        ),
        E.att(name_value("shared interaction", "interacts with")),
        E.att(
            name_value(
                "name",
                f"{G.nodes[node_id_1]['name']} (interacts with) {G.nodes[node_id_2]['name']}",
            )
        ),
        E.att(name_value("selected", "0", with_type="boolean")),
        E.att(name_value("interaction", "interacts with")),
        E.att(name_value("edge weight", str(edge_attr["edge_weight"]), with_type="double")),
        E.att(
            name_value(
                "scaled edge weight", str(edge_attr["scaled_edge_weight"]), with_type="double"
            )
        ),
        E.att(name_value("edge width", str(edge_attr["edge_width"]), with_type="integer")),
        E.att(name_value("pubmed id", ",".join(edge_attr["pmids"]))),
        E.graphics(
            _graphics_attr,
            # E.att(name_value("EDGE_TOOLTIP", "")),
            # E.att(name_value("EDGE_SELECTED", "false")),
            # E.att(name_value("EDGE_TARGET_ARROW_SIZE", "6.0")),
            # E.att(name_value("EDGE_LABEL", "")),
            # E.att(name_value("EDGE_LABEL_TRANSPARENCY", "255")),
            # E.att(name_value("EDGE_STACKING_DENSITY", "0.5")),
            # E.att(name_value("EDGE_TARGET_ARROW_SHAPE", "NONE")),
            # E.att(
            #     name_value("EDGE_SOURCE_ARROW_UNSELECTED_PAINT",
            #                "#000000")),
            # E.att(name_value("EDGE_TARGET_ARROW_SELECTED_PAINT",
            #                  "#FFFF00")),
            # E.att(
            #     name_value(
            #         "EDGE_TARGET_ARROW_UNSELECTED_PAINT",
            #         "#000000")),
            # E.att(name_value("EDGE_SOURCE_ARROW_SHAPE", "None")),
            # E.att(name_value("EDGE_BEND", "")),
            # E.att(name_value("EDGE_STACKING", "AUTO_BEND")),
            # E.att(name_value("EDGE_LABEL_COLOR", "#000000")),
            # E.att(name_value("EDGE_TRANSPARENCY", "255")),
            # E.att(name_value("EDGE_LABEL_ROTATION", "0.0")),
            # E.att(name_value("EDGE_LABEL_WIDTH", "200.0")),
            # E.att(name_value("EDGE_CURVED", "true")),
            # E.att(name_value("EDGE_SOURCE_ARROW_SIZE", "6.0")),
            # E.att(name_value("EDGE_VISIBLE", "true")),
            # E.att(name_value("EDGE_LINE_TYPE", "SOLID")),
            # E.att(name_value("EDGE_STROKE_SELECTED_PAINT", "#FF0000")),
            # E.att(name_value("EDGE_LABEL_FONT_SIZE", "10")),
            # E.att(
            #     name_value("EDGE_LABEL_FONT_FACE",
            #                "Dialog.plain,plain,10")),
            # E.att(name_value("EDGE_Z_ORDER", "0.0")),
            # E.att(name_value("EDGE_SOURCE_ARROW_SELECTED_PAINT",
            #                  "#FFFF00")),
        ),
    )

    return edge



from typing import Literal

import networkx as nx

from netmedex.network_core import NetworkBuilder


def filter_node(G: nx.Graph, node_degree_threshold: int):
    for node, degree in list(G.degree()):
        if degree < node_degree_threshold:
            G.remove_node(node)


def rebuild_graph(
    node_degree,
    cut_weight,
    format: Literal["xgmml", "html"],
    G=None,
    with_layout=False,
    graph_path=None,
):
    if G is None:
        G = NetworkBuilder.load_graph(graph_path)

    network_builder = NetworkBuilder(
        pubtator_filepath="",
        savepath=None,
        node_type="",
        output_filetype="",
        weighting_method="",
        edge_weight_cutoff=cut_weight,
        pmid_weight_filepath=None,
        community=G.graph.get("is_community", False),
        max_edges=G.graph.get("max_edges", 0),
        debug=False,
    )

    network_builder.remove_edges_by_weight(G)
    network_builder.remove_edges_by_rank(G)
    network_builder.remove_isolated_nodes(G)
    filter_node(G, node_degree)

    if with_layout:
        network_builder.set_network_layout(G)

    if G.graph.get("is_community", False) and format == "html":
        network_builder.set_network_communities(G)

    return G



import csv
import importlib
import logging
import math
import pickle
import re
from collections import defaultdict
from collections.abc import Sequence
from dataclasses import asdict
from datetime import datetime
from operator import itemgetter
from typing import Literal

import networkx as nx

from netmedex.graph_data import CommunityEdgeData, GraphEdgeData, GraphNodeData
from netmedex.npmi import normalized_pointwise_mutual_information
from netmedex.pubtator_parser import (
    PubTatorParser,
    PubTatorResult,
)
from netmedex.utils import generate_uuid

# VARIANT_PATTERN = re.compile(r"CorrespondingGene:.*CorrespondingSpecies:\d+")

MUTATION_PATTERNS = {
    "tmvar": re.compile(r"(tmVar:[^;]+)"),
    "hgvs": re.compile(r"(HGVS:[^;]+)"),
    "rs": re.compile(r"(RS#:[^;]+)"),
    "variantgroup": re.compile(r"(VariantGroup:[^;]+)"),
    "gene": re.compile(r"(CorrespondingGene:[^;]+)"),
    "species": re.compile(r"(CorrespondingSpecies:[^;]+)"),
    "ca": re.compile(r"(CA#:[^;]+)"),
}

NODE_COLOR_MAP = {
    "Chemical": "#67A9CF",
    "Gene": "#74C476",
    "Species": "#FD8D3C",
    "Disease": "#8C96C6",
    "DNAMutation": "#FCCDE5",
    "ProteinMutation": "#FA9FB5",
    "CellLine": "#BDBDBD",
    "SNP": "#FFFFB3",
}
NODE_SHAPE_MAP = {
    "Chemical": "ELLIPSE",
    "Gene": "TRIANGLE",
    "Species": "DIAMOND",
    "Disease": "ROUNDRECTANGLE",
    "DNAMutation": "PARALLELOGRAM",
    "ProteinMutation": "HEXAGON",
    "CellLine": "VEE",
    "SNP": "OCTAGON",
}

MIN_EDGE_WIDTH = 0
MAX_EDGE_WIDTH = 20
EDGE_BASE_COLOR = "#AD1A66"

# TODO: other ways to solve modified MeSH mismatch
mesh_map = {}
logger = logging.getLogger(__name__)


class NetworkBuilder:
    def __init__(
        self,
        pubtator_filepath: str,
        savepath: str | None,
        node_type: Literal["all", "mesh", "relation"],
        output_filetype: Literal["html", "xgmml", "json"],
        weighting_method: Literal["freq", "npmi"],
        edge_weight_cutoff: int,
        pmid_weight_filepath: str | None,
        community: bool,
        max_edges: int,
        debug: bool,
    ):
        self.pubtator_filepath = pubtator_filepath
        self.savepath = savepath
        self.node_type = node_type
        self.output_filetype = output_filetype
        self.weighting_method = weighting_method
        self.edge_weight_cutoff = edge_weight_cutoff
        self.pmid_weight_filepath = pmid_weight_filepath
        self.community = community
        self.max_edges = max_edges
        self.debug = debug

    def run(self):
        self.check_not_implemented()
        G = nx.Graph()
        result = PubTatorParser(
            pubtator_filepath=self.pubtator_filepath,
            node_type=self.node_type,
        ).parse()
        self.add_node_to_graph(G=G, parsed_pubtator=result)
        self.add_edge_to_graph(G=G, parsed_pubtator=result)
        self.add_attrs_to_graph(G=G, parsed_pubtator=result)

        if self.debug:
            # Save before truncation
            now = datetime.now().strftime("%y%m%d%H%M%S")
            with open(f"tocytoscape_graph_{now}.pkl", "wb") as f:
                pickle.dump(G, f)

        self.remove_edges_by_weight(G)
        self.remove_edges_by_rank(G)

        self.remove_isolated_nodes(G)

        self.assert_graph_properties(G)

        self.set_network_layout(G)

        if self.community:
            self.set_network_communities(G)

        self.log_graph_info(G, result)
        self.save_network(G)

        return G

    def check_not_implemented(self):
        if self.community and self.output_filetype == "xgmml":
            msg = "Save community in XGMML format not yet supported."
            logger.info(msg)
            raise NotImplementedError(msg)

    @staticmethod
    def add_node_to_graph(G: nx.Graph, parsed_pubtator: PubTatorResult):
        # TODO: add feature: mark specific names
        marked = False
        for node_id in parsed_pubtator.non_isolated_nodes:
            node_data = GraphNodeData(
                _id=str(parsed_pubtator.node_dict[node_id].id),
                color=NODE_COLOR_MAP[parsed_pubtator.node_dict[node_id].type],
                label_color="#000000",
                shape=NODE_SHAPE_MAP[parsed_pubtator.node_dict[node_id].type],
                type=parsed_pubtator.node_dict[node_id].type,
                name=parsed_pubtator.node_dict[node_id].name,
                document_frequency=len(parsed_pubtator.node_dict[node_id].pmids),
                marked=marked,
            )
            G.add_node(node_id, **asdict(node_data))

    def add_edge_to_graph(self, G: nx.Graph, parsed_pubtator: PubTatorResult):
        doc_weights = {}
        if self.pmid_weight_filepath is not None:
            with open(self.pmid_weight_filepath, newline="") as f:
                reader = csv.reader(f)
                for row in reader:
                    doc_weights[row[0]] = float(row[1])

        for pair, edge_data_list in parsed_pubtator.edge_dict.items():
            if not G.has_node(pair[0]) or not G.has_node(pair[1]):
                continue

            pmids = [edge_data.pmid for edge_data in edge_data_list]
            unique_pmids = set(pmids)

            w_freq = round(sum([doc_weights.get(pmid, 1) for pmid in unique_pmids]), 2)
            npmi = normalized_pointwise_mutual_information(
                n_x=len(parsed_pubtator.node_dict[pair[0]].pmids),
                n_y=len(parsed_pubtator.node_dict[pair[1]].pmids),
                n_xy=len(unique_pmids),
                N=parsed_pubtator.num_pmids,
                n_threshold=2,
                below_threshold_default=MIN_EDGE_WIDTH / MAX_EDGE_WIDTH,
            )

            if self.weighting_method == "npmi":
                edge_weight = npmi
            elif self.weighting_method == "freq":
                edge_weight = w_freq

            edge_data = GraphEdgeData(
                _id=str(edge_data_list[0].id),
                raw_frequency=len(unique_pmids),
                doc_weighted_frequency=w_freq,
                npmi=npmi,
                edge_weight=edge_weight,
                pmids=list(unique_pmids),
            )

            G.add_edge(pair[0], pair[1], type="node", **asdict(edge_data))

        edge_weights = nx.get_edge_attributes(G, "edge_weight")
        scale_factor = self._calculate_scale_factor(
            edge_weights=edge_weights.values(),
            max_width=MAX_EDGE_WIDTH,
        )
        for edge, weight in edge_weights.items():
            scaled_weight = max(weight * scale_factor, 0.0)
            G.edges[edge]["scaled_edge_weight"] = round(scaled_weight, 2)
            G.edges[edge]["edge_width"] = int(max(scaled_weight, MIN_EDGE_WIDTH))

    def _calculate_scale_factor(
        self,
        edge_weights: Sequence[float],
        max_width: int,
    ):
        max_weight = max(edge_weights)
        if self.weighting_method == "npmi":
            scale_factor = max_width
        elif self.weighting_method == "freq":
            scale_factor = min(max_width / max_weight, 1)

        return scale_factor

    def remove_edges_by_weight(self, G: nx.Graph):
        to_remove = []
        for u, v, edge_attrs in G.edges(data=True):
            if edge_attrs["edge_width"] < self.edge_weight_cutoff:
                to_remove.append((u, v))
        G.remove_edges_from(to_remove)

    def remove_edges_by_rank(self, G: nx.Graph):
        if self.max_edges <= 0:
            return

        if G.number_of_edges() > self.max_edges:
            edges = sorted(
                G.edges(data=True), key=lambda x: x[2]["scaled_edge_weight"], reverse=True
            )
            for edge in edges[self.max_edges :]:
                G.remove_edge(edge[0], edge[1])

    @staticmethod
    def remove_isolated_nodes(G: nx.Graph):
        G.remove_nodes_from(list(nx.isolates(G)))

    @staticmethod
    def add_attrs_to_graph(G: nx.Graph, parsed_pubtator: PubTatorResult):
        G.graph["pmid_title"] = parsed_pubtator.pmid_title_map

    @staticmethod
    def assert_graph_properties(G: nx.Graph):
        num_selfloops = nx.number_of_selfloops(G)
        assert num_selfloops == 0, f"Find {num_selfloops}"

    @staticmethod
    def set_network_layout(G: nx.Graph):
        if G.number_of_edges() > 1000:
            pos = nx.circular_layout(G, scale=300)
        else:
            pos = nx.spring_layout(
                G, weight="scaled_edge_weight", scale=300, k=0.25, iterations=15
            )
        nx.set_node_attributes(G, pos, "pos")

    @staticmethod
    def set_network_communities(G: nx.Graph, seed: int = 1):
        communities = nx.community.louvain_communities(G, seed=seed, weight="scaled_edge_weight")
        community_labels = set()
        for c_idx, community in enumerate(communities):
            highest_degree_node = max(
                G.degree(community, weight="scaled_edge_weight"), key=itemgetter(1)
            )[0]
            community_node = f"c{c_idx}"
            community_labels.add(community_node)
            community_attrs = G.nodes[highest_degree_node].copy()
            community_attrs.update(
                {"label_color": "#dd4444", "parent": None, "_id": community_node}
            )
            node_data = GraphNodeData(**community_attrs)
            G.add_node(community_node, **asdict(node_data))

            for node in community:
                G.nodes[node]["parent"] = community_node

        G.graph["num_communities"] = len(community_labels)

        # Gather edges between communities
        inter_edge_weight = defaultdict(int)
        inter_edge_pmids = defaultdict(list)
        to_remove = []
        for u, v, attrs in G.edges(data=True):
            if (c_0 := G.nodes[u]["parent"]) != (c_1 := G.nodes[v]["parent"]):
                assert c_0 is not None and c_1 is not None
                to_remove.append((u, v))
                community_edge = tuple(sorted([c_0, c_1]))
                inter_edge_weight[community_edge] += attrs["scaled_edge_weight"]
                inter_edge_pmids[community_edge].extend(attrs["pmids"])

        G.remove_edges_from(to_remove)
        for (c_0, c_1), weight in inter_edge_weight.items():
            # Log-adjusted weight for balance
            try:
                weight = math.log(weight) * 5
                weight = 0.0 if weight < 0.0 else weight
            except ValueError:
                weight = 0.0
            pmids = list(set(inter_edge_pmids[(c_0, c_1)]))
            edge_data = CommunityEdgeData(
                _id=generate_uuid(),
                edge_weight=weight,
                scaled_edge_weight=weight,
                edge_width=int(max(weight, MIN_EDGE_WIDTH)),
                pmids=pmids,
            )
            G.add_edge(c_0, c_1, type="community", **asdict(edge_data))

    @staticmethod
    def log_graph_info(G: nx.Graph, parsed_pubtator: PubTatorResult):
        logger.info(f"# articles: {parsed_pubtator.num_pmids}")
        if num_communities := G.graph.get("num_communities", 0):
            logger.info(f"# communities: {num_communities}")
        logger.info(f"# nodes: {G.number_of_nodes() - num_communities}")
        logger.info(f"# edges: {G.number_of_edges()}")

    def save_network(self, G: nx.Graph):
        if self.savepath is None:
            return
        FORMAT_FUNCTION_MAP = {
            "xgmml": "netmedex.cytoscape_xgmml.save_as_xgmml",
            "html": "netmedex.cytoscape_js.save_as_html",
            "json": "netmedex.cytoscape_js.save_as_json",
        }

        module_path, func_name = FORMAT_FUNCTION_MAP[self.output_filetype].rsplit(".", 1)
        module = importlib.import_module(module_path)
        save_func = getattr(module, func_name)

        save_func(G, self.savepath)
        logger.info(f"Save graph to {self.savepath}")

    @staticmethod
    def load_graph(graph_pickle_path: str):
        with open(graph_pickle_path, "rb") as f:
            G = pickle.load(f)

        return G



import dash
import dash_bootstrap_components as dbc
from dash import dcc, html

from webapp.components.utils import generate_param_title
from webapp.utils import visibility

max_articles = html.Div(
    [
        generate_param_title(
            "Max Articles",
            "Set the maximum number of articles retrieved via API",
        ),
        dcc.Slider(
            50,
            3000,
            50,
            value=1000,
            marks=None,
            id="max-articles",
            tooltip={"placement": "bottom", "always_visible": False},
        ),
    ],
    className="param",
)

max_edges = html.Div(
    [
        generate_param_title(
            "Max Edges",
            "Set the maximum number of edges to display in the graph (0: No limit)",
        ),
        dcc.Slider(
            0,
            1000,
            50,
            value=1000,
            marks=None,
            id="max-edges",
            tooltip={"placement": "bottom", "always_visible": False},
        ),
    ],
    className="param",
)


advanced_settings = html.Div(
    [
        dbc.Button(
            html.Img(src=dash.get_asset_url("icon_config.svg"), width=22, height=22),
            # "Settings",
            id="advanced-settings-btn",
            className="btn-secondary settings",
        ),
        html.Div(
            [
                html.H5("Advanced Settings", className="text-center"),
                max_articles,
                max_edges,
            ],
            id="advanced-settings-collapse",
            className="settings-collapse",
            style=visibility.hidden,
        ),
    ],
    id="advanced-settings",
    className="settings-container",
)



import base64

from dash import Input, Output, State, html, no_update


def display_uploaded_data(data, filename):
    if data is not None:
        content_type, content_string = data.split(",")
        decoded_content = base64.b64decode(content_string).decode("utf-8")
        displayed_text = decoded_content.split("\n")[:5]
        displayed_text = [t[:100] + "..." if len(t) > 100 else t for t in displayed_text]
        return [
            html.H6(f"File: {filename}", style={"marginBottom": "5px", "marginTop": "5px"}),
            html.Pre("\n".join(displayed_text), className="upload-preview"),
        ]
    else:
        return no_update


def callbacks(app):
    @app.callback(
        Output("output-data-upload", "children"),
        Input("pmid-file-data", "contents"),
        State("pmid-file-data", "filename"),
    )
    def update_data_upload(upload_data, filename):
        return display_uploaded_data(upload_data, filename)

    @app.callback(
        Output("pubtator-file-upload", "children"),
        Input("pubtator-file-data", "contents"),
        State("pubtator-file-data", "filename"),
    )
    def update_pubtator_upload(pubtator_data, filename):
        return display_uploaded_data(pubtator_data, filename)



from dash import Input, Output, State

from webapp.utils import visibility


def callbacks(app):
    @app.callback(
        Output("graph-settings-collapse", "style", allow_duplicate=True),
        Output("graph-cut-weight", "tooltip"),
        Input("graph-settings-btn", "n_clicks"),
        State("graph-settings-collapse", "style"),
        prevent_initial_call=True,
    )
    def open_graph_settings(n_clicks, style):
        visibility = style["visibility"]
        toggle = {"hidden": "visible", "visible": "hidden"}
        weight_toggle = {"hidden": True, "visible": False}
        return (
            {"visibility": toggle[visibility]},
            {"placement": "bottom", "always_visible": weight_toggle[visibility]},
        )

    @app.callback(
        Output("graph-settings-collapse", "style", allow_duplicate=True),
        Output("graph-cut-weight", "value"),
        Output("graph-cut-weight", "tooltip", allow_duplicate=True),
        Output("download-pubtator-btn", "style"),
        Input("cy-graph-container", "style"),
        State("memory-graph-cut-weight", "data"),
        State("api-toggle-items", "value"),
        prevent_initial_call=True,
    )
    def update_graph_params(container_style, cut_weight, api_or_file):
        if api_or_file == "api" and container_style.get("visibility") == "visible":
            pubtator_btn_visibility = visibility.visible
        else:
            pubtator_btn_visibility = visibility.hidden
        return (
            visibility.hidden,
            cut_weight,
            {"placement": "bottom", "always_visible": False},
            pubtator_btn_visibility,
        )



from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

import pytest

from netmedex.network_core import NetworkBuilder
from netmedex.pubtator_parser import (
    HEADER_SYMBOL,
    PubTatorEdgeData,
    PubTatorLine,
    PubTatorNodeData,
    PubTatorParser,
)


@pytest.fixture(scope="module")
def tempdir():
    with TemporaryDirectory() as tempdir:
        yield Path(tempdir)


@pytest.fixture(scope="module")
def paths():
    return {
        "simple": "./tests/test_data/6_nodes_3_clusters_mesh.pubtator",
        "mesh_collision": "./tests/test_data/mesh_collision.pubtator",
        "merge_genes": "./tests/test_data/merge_genes.pubtator",
    }


@pytest.fixture()
def network_args():
    return {
        "pubtator_filepath": None,
        "savepath": None,
        "node_type": "all",
        "output_filetype": "html",
        "weighting_method": "freq",
        "edge_weight_cutoff": 0,
        "pmid_weight_filepath": None,
        "community": False,
        "max_edges": 0,
        "debug": False,
    }


def test_index_by_text(paths, network_args):
    network_args["pubtator_filepath"] = paths["simple"]
    network_builder = NetworkBuilder(**network_args)
    G = network_builder.run()

    assert G.nodes.get(
        "RS#:854560;HGVS:p.L55M;CorrespondingGene:5444", False
    ), "l55m should be in the graph"
    assert len(G.nodes) == 14, f"result: {len(G.nodes)} nodes\nexpected: 14 nodes"
    assert len(G.edges) == 47, f"result: {len(G.edges)} nodes\nexpected: 47 edges"


def test_index_by_relation(paths, network_args):
    network_args["pubtator_filepath"] = paths["simple"]
    network_args["node_type"] = "relation"
    network_builder = NetworkBuilder(**network_args)
    G = network_builder.run()

    assert G.nodes.get(
        "RS#:854560;HGVS:p.L55M;CorrespondingGene:5444", False
    ), "Not mapped by MeSH"
    assert len(G.nodes) == 8, f"result: {len(G.nodes)} nodes\nexpected: 8 nodes"
    assert len(G.edges) == 9, f"result: {len(G.edges)} nodes\nexpected: 9 edges"


def test_index_by_mesh(paths, network_args):
    network_args["pubtator_filepath"] = paths["simple"]
    network_args["node_type"] = "mesh"
    network_builder = NetworkBuilder(**network_args)
    G = network_builder.run()

    assert G.nodes.get(
        "RS#:854560;HGVS:p.L55M;CorrespondingGene:5444", False
    ), "Not mapped by MeSH"
    assert len(G.nodes) == 13, f"result: {len(G.nodes)} nodes\nexpected: 13 nodes"
    assert len(G.edges) == 41, f"result: {len(G.edges)} nodes\nexpected: 41 edges"


def test_mesh_collision_handling(paths, network_args):
    network_args["pubtator_filepath"] = paths["mesh_collision"]
    network_builder = NetworkBuilder(**network_args)
    G = network_builder.run()

    assert len(G.nodes) == 7, f"result: {len(G.nodes)} nodes\nexpected: 7 nodes"
    assert len(G.edges) == 21, f"result: {len(G.edges)} nodes\nexpected: 21 edges"


def test_merge_genes(paths, network_args):
    network_args["pubtator_filepath"] = paths["merge_genes"]
    network_builder = NetworkBuilder(**network_args)
    G = network_builder.run()

    assert len(G.nodes) == 7, f"result: {len(G.nodes)} nodes\nexpected: 7 nodes"
    assert len(G.edges) == 11, f"result: {len(G.nodes)} nodes\nexpected: 11 edges"


@pytest.mark.parametrize(
    "data,expected",
    [
        (f"{HEADER_SYMBOL}USE-MESH-VOCABULARY\nfoobar", True),
        (f"{HEADER_SYMBOL[0]}USE-MESH-VOCABULARY\nfoobar", False),
        (f"{HEADER_SYMBOL + HEADER_SYMBOL}USE-MESH-VOCABULARY\nfoobar", False),
    ],
)
def test_parse_header(data, expected, monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("builtins.open", mock.mock_open(read_data=data))
    pubtator_line = PubTatorLine
    with mock.patch("netmedex.pubtator_parser.PubTatorLine", pubtator_line):
        parser = PubTatorParser(data, node_type="all")
        parser.parse()
        assert pubtator_line.use_mesh == expected


def test_merge_same_name_genes():
    node_dict = {
        "gene_1": PubTatorNodeData(id=1, mesh="1", type="Gene", name="foo", pmids={10}),
        "gene_2": PubTatorNodeData(id=2, mesh="2", type="Gene", name="foo", pmids={20}),
        "bar": PubTatorNodeData(id=3, mesh="-", type="Chemical", name="bar", pmids={30}),
    }
    edge_dict = {
        ("gene_1", "bar"): [PubTatorEdgeData(id=4, pmid="30"), PubTatorEdgeData(id=5, pmid="40")],
        ("gene_1", "gene_2"): [PubTatorEdgeData(id=6, pmid="20")],
        ("foo", "bar"): [PubTatorEdgeData(id=7, pmid="30"), PubTatorEdgeData(id=8, pmid="40")],
    }
    parser = PubTatorParser(pubtator_filepath="", node_type="all")
    parser.node_dict = node_dict
    parser.edge_dict = edge_dict
    parser._merge_same_name_genes()

    assert node_dict == {
        "bar": PubTatorNodeData(id=3, mesh="-", type="Chemical", name="bar", pmids={30}),
        "gene_1;2": PubTatorNodeData(id=2, mesh="1;2", type="Gene", name="foo", pmids={10, 20}),
    }
    assert edge_dict == {
        ("gene_1;2", "bar"): [
            PubTatorEdgeData(id=4, pmid="30"),
            PubTatorEdgeData(id=5, pmid="40"),
        ],
        ("foo", "bar"): [PubTatorEdgeData(id=7, pmid="30"), PubTatorEdgeData(id=8, pmid="40")],
    }


def test_save_network(tempdir, paths, network_args):
    network_args["pubtator_filepath"] = paths["simple"]
    network_args["savepath"] = str(tempdir / "output.html")
    NetworkBuilder(**network_args).run()



from dash import Input, Output, State, dcc

from netmedex.cytoscape_js import save_as_html
from netmedex.cytoscape_xgmml import save_as_xgmml
from webapp.callbacks.graph_utils import rebuild_graph


def callbacks(app):
    @app.callback(
        Output("download-pubtator", "data"),
        Input("download-pubtator-btn", "n_clicks"),
        State("current-session-path", "data"),
        prevent_initial_call=True,
    )
    def download_pubtator(n_clicks, savepath):
        return dcc.send_file(savepath["pubtator"], filename="output.pubtator")

    @app.callback(
        Output("export-html", "data"),
        Input("export-btn-html", "n_clicks"),
        State("graph-layout", "value"),
        State("node-degree", "value"),
        State("graph-cut-weight", "value"),
        State("current-session-path", "data"),
        prevent_initial_call=True,
    )
    def export_html(n_clicks, layout, node_degree, weight, savepath):
        G = rebuild_graph(
            node_degree, weight, format="html", with_layout=True, graph_path=savepath["graph"]
        )
        save_as_html(G, savepath["html"], layout=layout)
        return dcc.send_file(savepath["html"], filename="output.html")

    @app.callback(
        Output("export-xgmml", "data"),
        Input("export-btn-xgmml", "n_clicks"),
        State("graph-layout", "value"),
        State("node-degree", "value"),
        State("graph-cut-weight", "value"),
        State("current-session-path", "data"),
        prevent_initial_call=True,
    )
    def export_xgmml(n_clicks, layout, node_degree, weight, savepath):
        G = rebuild_graph(
            node_degree, weight, format="xgmml", with_layout=True, graph_path=savepath["graph"]
        )
        save_as_xgmml(G, savepath["xgmml"])
        return dcc.send_file(savepath["xgmml"], filename="output.xgmml")

    @app.callback(
        Output("export-edge-csv", "data"),
        Input("export-edge-btn", "n_clicks"),
        State("cy", "tapEdgeData"),
        State("pmid-title-dict", "data"),
        State("current-session-path", "data"),
        prevent_initial_call=True,
    )
    def export_edge_csv(n_clicks, tap_edge, pmid_title, savepath):
        import csv

        with open(savepath["edge_info"], "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerow(["PMID", "Title"])
            writer.writerows([[pmid, pmid_title[pmid]] for pmid in tap_edge["pmids"]])
        n1, n2 = tap_edge["label"].split(" (interacts with) ")
        filename = f"{n1}_{n2}.csv"
        return dcc.send_file(savepath["edge_info"], filename=filename)



import dash
import dash_bootstrap_components as dbc
from dash import dcc, html

from webapp.components.utils import icon_download
from webapp.utils import visibility


def create_legend_box(icon, text):
    return html.Div(
        [
            html.Img(src=dash.get_asset_url(icon), width=25, height=25),
            html.P(text),
        ],
        className="legend-box",
    )


edge_info = html.Div(
    [
        html.H5("Edge Info", className="text-center"),
        dbc.Button([icon_download(), "CSV"], id="export-edge-btn", className="export-btn"),
        dcc.Download(id="export-edge-csv"),
        html.Div(id="edge-info"),
    ],
    id="edge-info-container",
    className="flex-grow-1",
    style=visibility.hidden,
)


legend = html.Div(
    [
        create_legend_box("icon_species.svg", "Species"),
        create_legend_box("icon_chemical.svg", "Chemical"),
        create_legend_box("icon_gene.svg", "Gene"),
        create_legend_box("icon_disease.svg", "Disease"),
        create_legend_box("icon_cellline.svg", "CellLine"),
        create_legend_box("icon_dnamutation.svg", "DNAMutation"),
        create_legend_box("icon_proteinmutation.svg", "ProteinMutation"),
        create_legend_box("icon_snp.svg", "SNP"),
    ],
    id="legend-container",
)

graph_info = html.Div([edge_info, legend], id="bottom-container", className="d-flex")
